include "../../include/config.pxi"

IF USE_WAYLAND:
    cdef class WindowInfoWayland:
        @property
        def display(self):
            return <uintptr_t>self.display

        @property
        def surface(self):
            return <uintptr_t>self.surface

        @property
        def shell_surface(self):
            return <uintptr_t>self.shell_surface


IF USE_X11:
    cdef class WindowInfoX11:
        @property
        def display(self):
            return <uintptr_t>self.display

        @property
        def window(self):
            return <uintptr_t>self.window

IF UNAME_SYSNAME == 'Windows':
    cdef class WindowInfoWindows:
        @property
        def window(self):
            return <uintptr_t>self.window

        @property
        def hdc(self):
            return <uintptr_t>self.hdc



'''
BoxShadow
=========

.. versionadded:: 2.2.0


BoxShadow is a graphical instruction used to add a shadow effect to an element.

Its behavior is similar to the concept of a CSS3 box-shadow.

.. image:: images/boxshadow.png
    :align: center

The BoxShadow declaration must occur inside a :class:`~kivy.graphics.instructions.Canvas` statement. It works
similarly to other graphical instructions such as :class:`~kivy.graphics.vertex_instructions.Rectangle`,
:class:`~kivy.graphics.vertex_instructions.RoundedRectangle`, etc.

.. note::

    Although the ``BoxShadow`` graphical instruction has a visually similar behavior to box-shadow (CSS), the hierarchy
    of the drawing layer of ``BoxShadow`` in relation to the target element must be defined following the same layer
    hierarchy rules as when declaring other canvas instructions.
    
    |

    For more details, refer to the :attr:`~kivy.graphics.boxshadow.BoxShadow.inset` mode.

.. _example:

Example:
--------

    .. image:: images/boxshadow_demo.gif
        :align: center

    .. code-block:: kv

        <MyWidget>:
            Button:
                pos_hint: {"center_x": 0.5, "center_y": 0.5}
                size_hint: None, None
                size: 200, 150
                background_down: self.background_normal
                canvas.before:
                    Color:
                        rgba: 0, 0, 1, 0.85
                    BoxShadow:
                        pos: self.pos
                        size: self.size
                        offset: 0, -10
                        spread_radius: -20, -20
                        border_radius: 10, 10, 10, 10
                        blur_radius: 80 if self.state == "normal" else 50

'''

__all__ = ("BoxShadow", )

from kivy.logger import Logger
from kivy.graphics.fbo cimport Fbo
from kivy.graphics.vertex_instructions import Rectangle, RoundedRectangle
from kivy.graphics.context_instructions cimport Translate, Scale
from kivy.graphics.gl_instructions import ClearBuffers, ClearColor
from kivy.graphics.instructions cimport InstructionGroup


cdef str SHADOW_fs = """
#ifdef GL_ES
    precision highp float;
#endif

/* Outputs from the vertex shader */
varying vec4 frag_color;
varying vec2 tex_coord0;
/* uniform texture samplers */
uniform sampler2D texture0;

uniform int inset;
uniform float blur_radius;
uniform vec4 border_radius;
uniform vec2 spread_radius;
uniform vec2 size;
uniform vec2 offset;

// References:
// https://www.shadertoy.com/view/3tj3Dm
// https://www.shadertoy.com/view/NtVSW1
// https://iquilezles.org/articles/distfunctions2d/

float roundedBoxSDF(vec2 pos, vec2 size, vec4 radius){
    radius = min(radius, min(size.x, size.y));
    vec2 s = step(pos, vec2(0.0));
    float r = mix(mix(radius.y, radius.z, s.y), mix(radius.x, radius.w, s.y), s.x);
    vec2 q = abs(pos) - size + r;
    return min(max(q.x, q.y), 0.0) + length(max(q,0.0)) - r;
}

float sigmoid(float x) {
    return 1.0 / (1.0 + exp(-x));
}

void main (void){

    float adjustmentFactor = 0.0;
    vec2 boxSize = size / 2.0;
    vec2 boxPos = tex_coord0 * size - size/2.0;

    if (inset == 1){
        boxPos -= offset;
        boxSize -= blur_radius + spread_radius;
        adjustmentFactor = blur_radius;
    }
    else{
        boxSize -= blur_radius * 1.5 + vec2(2.0);
    }

    float distShadow = sigmoid(
        (
            roundedBoxSDF(
                boxPos,
                boxSize,
                border_radius - adjustmentFactor
            ) - adjustmentFactor
        ) / (max(1.0, blur_radius) / 4.0)
    );

    if (inset == 1){
        gl_FragColor = vec4(1.0, 1.0, 1.0, distShadow);
    }
    else{
        gl_FragColor = vec4(1.0, 1.0, 1.0, (1.0 - distShadow) * (frag_color.a * 2.0));
    }

}
"""


cdef class BoxShadow(InstructionGroup):

    '''A box shadow effect.

    .. versionadded:: 2.2.0

    .. versionchanged:: 2.3.0
        Fixed :class:`~kivy.graphics.instructions.Canvas` management using
        :meth:`~kivy.graphics.instructions.Canvas.add`,
        :meth:`~kivy.graphics.instructions.Canvas.insert` and
        :meth:`~kivy.graphics.instructions.Canvas.remove`.
        Previously, using them to manage the :class:`~kivy.graphics.instructions.Canvas` had no effect.

        The base class also changed from :class:`~kivy.graphics.fbo.Fbo`
        to :class:`~kivy.graphics.instructions.InstructionGroup`.


    :Parameters:

        `inset`: bool, defaults to ``False``.
            Defines whether the shadow is drawn from the inside out or from the
            outline to the inside of the ``BoxShadow`` instruction.
        `size`: list | tuple, defaults to ``(100.0, 100.0)``.
            Define the raw size of the shadow, that is, you should not take into account
            changes in the value of :attr:`blur_radius` and :attr:`spread_radius`
            properties when setting this parameter.
        `pos`: list | tuple, defaults to ``(0.0, 0.0)``.
            Define the raw position of the shadow, that is, you should not take into account
            changes in the value of the :attr:`offset` property when setting this parameter.
        `offset`: list | tuple, defaults to ``(0.0, 0.0)``.
            Specifies shadow offsets in `(horizontal, vertical)` format. 
            Positive values for the offset indicate that the shadow should move to the right and/or top.
            The negative ones indicate that the shadow should move to the left and/or down.
        `blur_radius`: float, defaults to ``15.0``.
            Define the shadow blur radius. Controls shadow expansion and softness.
        `spread_radius`: list | tuple, defaults to ``(0.0, 0.0)``.
            Define the shrink/expansion of the shadow.
        `border_radius`: list | tuple, defaults to ``(0.0, 0.0, 0.0, 0.0)``.
            Specifies the radii used for the rounded corners clockwise:
            top-left, top-right, bottom-right, bottom-left.
    '''

    def __init__(self, *args, **kwargs):
        super().__init__()
        inset = kwargs.get("inset", False)
        pos = kwargs.get("pos", (0.0, 0.0))
        size = kwargs.get("size", (100.0, 100.0))
        offset = kwargs.get("offset", (0.0, 0.0))
        blur_radius = kwargs.get("blur_radius", 15.0)
        spread_radius = kwargs.get("spread_radius", (0.0, 0.0))
        border_radius = kwargs.get("border_radius", (0.0, 0.0, 0.0, 0.0))

        self._inset = self._check_bool(inset)
        self._pos = self._check_iter("pos", pos)
        self._size = self._bounded_value(
            self._check_iter("size", size),
            min_value=0.0
        )
        self._offset = self._check_iter("offset", offset)
        self._blur_radius = self._bounded_value(
            self._check_float("blur_radius", blur_radius),
            min_value=0.0
        )
        self._spread_radius = self._check_iter("spread_radius", spread_radius)
        self._border_radius = self._bounded_value(
            self._check_iter("border_radius", border_radius, components=4),
            min_value=1.0,
            max_value=min(self.size)
        )
        self._init_texture()

    cdef void _init_texture(self):
        self._fbo = Fbo(noadd=True, size=(100, 100), fs=SHADOW_fs)
        self.add(self._fbo)
        with self._fbo:
            ClearColor(1, 1, 1, 0)
            ClearBuffers()
            self._fbo_translate = Translate(0, 0)
            self._fbo_scale = Scale(1, 1, 1)
            self._fbo_rect = Rectangle(size=(100, 100))
        self._texture_container = RoundedRectangle(
            noadd=True,
            texture=self._fbo.texture,
            size=(100, 100),
            radius=(1, 1, 1, 1),
            segments=45
        )
        self.add(self._texture_container)
        self._update_shadow()

    cdef void _update_canvas(self):
        self._texture_container.pos = self._fbo_rect.pos = self.pos
        self._texture_container.size = self._fbo_rect.size = self.size
        if self.inset:
            self._texture_container.radius = self.border_radius

    cdef void _update_fbo(self):
        cdef float scale_x, scale_y
        if 0 in self.size:
            return
        scale_x = 100 / self.size[0]
        scale_y = 100 / self.size[1]
        self._fbo_scale.x = scale_x
        self._fbo_scale.y = scale_y
        self._fbo_translate.x = - scale_x * self.pos[0]
        self._fbo_translate.y = - scale_y * self.pos[1]

    cdef void _update_shadow(self):
        self._update_canvas()
        self._update_fbo()

        self._fbo["inset"] = int(self.inset)
        self._fbo["blur_radius"] = self.blur_radius
        self._fbo["border_radius"] = self.border_radius
        self._fbo["spread_radius"] = self.spread_radius
        self._fbo["offset"] = self.offset
        self._fbo["size"] = self.size

    cdef tuple _adjusted_pos(self):
        """Return the adjusted position of the shadow texture containers,
        based on the adjusted size and offset. The pos adjustment is only
        applied if inset is disabled.
        """
        cdef float x, y
        x, y = self._pos

        # The position should be adjusted according to the size expansion,
        # with half the size used in the _adjusted_size method
        if not self.inset:
            x -= self.blur_radius * 1.5 + self.spread_radius[0] - self.offset[0]
            y -= self.blur_radius * 1.5 + self.spread_radius[1] - self.offset[1]

        return (x, y)

    cdef tuple _adjusted_size(self):
        """Returns the adjusted size of the shadow texture containers, to avoid
        unwanted shadow cropping effect. The size adjustment is only applied if
        inset is disabled.
        """
        cdef float w, h
        w, h = self._size

        # size expansion
        if not self.inset and w > 0.0 and h > 0.0:
            w += self.blur_radius * 3 + self.spread_radius[0] * 2
            h += self.blur_radius * 3 + self.spread_radius[1] * 2

        w = max(0.0, w)
        h = max(0.0, h)
        return (w, h)

    cdef object _bounded_value(self, object value, min_value=None, max_value=None):
        cdef _value = []
        if isinstance(value, (list, tuple)):
            for v in value:
                if min_value is not None and v < min_value:
                    _value.append(min_value)
                elif max_value is not None and v > max_value:
                    _value.append(max_value)
                else:
                    _value.append(v)
            value = tuple(_value)

        elif isinstance(value, (int, float)):
            if min_value is not None and value < min_value:
                value = min_value
            elif max_value is not None and value > max_value:
                value = max_value
        return value

    cdef bint _check_bool(self, object value):
        if not isinstance(value, bool):
            raise TypeError(
                f"'inset' accept only boolean values (True or False), got {type(value)}."
            )
        return value

    cdef float _check_float(self, str property_name, object value, str iter_text=""):
        if not isinstance(value, (int, float)):
            raise TypeError(
                f"'{property_name}' accept only {iter_text} int/float got {type(value)}"
            )
        return float(value)

    cdef tuple _check_iter(self, str property_name, object value, int components=2):
        cdef int _len
        cdef list _value = []
        if not isinstance(value, (list, tuple)):
            raise TypeError(
                f"'{property_name}' accepts only list/tuple, got {type(value)}"
            )
        _len = len(value)
        if _len != components:
            raise ValueError(
                f"'{property_name}' must have {components}, got {_len}"
            )

        for v in value:
            _value.append(
                self._check_float(property_name, v, iter_text="list/tuple of")
            )
        return tuple(_value)

    @property
    def pos(self):
        '''Define the raw position of the shadow, that is, you should not take
        into account changes in the value of the :attr:`offset` property when
        setting this property.

        - :attr:`inset` **OFF**:
            Returns the adjusted position of the shadow according to the
            adjusted :attr:`size` of the shadow and :attr:`offset` property.

        - :attr:`inset` **ON**:
            Returns the raw position (the same as specified).

        Defaults to ``(0.0, 0.0)``.

        .. note::

            It is recommended that this property matches the raw position of
            the shadow target element. To manipulate horizontal and vertical
            offset, use :attr:`offset` instead.

        '''
        return self._adjusted_pos()

    @pos.setter
    def pos(self, value):
        self._pos = self._check_iter("pos", value)
        self._update_shadow()

    @property
    def size(self):
        '''Define the raw size of the shadow, that is, you should not take into
        account changes in the value of :attr:`blur_radius` and :attr:`spread_radius` properties.

        - :attr:`inset` **OFF**:
            Returns the adjusted size of the shadow according to the
            :attr:`blur_radius` and :attr:`spread_radius` properties.

        - :attr:`inset` **ON**:
            Returns the raw size (the same as specified).

        Defaults to ``(100.0, 100.0)``.

        .. note::

            It is recommended that this property matches the raw size of
            the shadow target element. To control the shrink/expansion of
            the shadow's raw :attr:`size`, use :attr:`spread_radius` instead.

        '''
        return self._adjusted_size()

    @size.setter
    def size(self, value):
        _size = self._check_iter("size", value)
        self._size = self._bounded_value(_size, min_value=0.0)
        self._update_shadow()

    @property
    def border_radius(self):
        '''Specifies the radii used for the rounded corners clockwise:
        top-left, top-right, bottom-right, bottom-left.

        Defaults to ``(0.0, 0.0, 0.0, 0.0)``.

        - :attr:`inset` **OFF**:
            .. image:: images/boxshadow_border_radius.svg
                :align: center
        
        | 

        - :attr:`inset` **ON**:
            .. image:: images/boxshadow_border_radius_inset.svg
                :align: center

        '''
        return self._bounded_value(self._border_radius, min_value=1.0, max_value=max(min(self.size) / 2, 1.0))

    @border_radius.setter
    def border_radius(self, value):
        self._border_radius = self._check_iter("border_radius", value, components=4)
        self._update_shadow()

    @property
    def spread_radius(self):
        '''Define the shrink/expansion of the shadow in `[horizontal, vertical]` format.

        Defaults to ``(0.0, 0.0)``.

        This property is especially useful for cases where you want to achieve
        a softer shadow around the element, by setting negative values for
        :attr:`spread_radius` and a larger value for :attr:`blur_radius` as
        in the :ref:`example <example>`.

        - :attr:`inset` **OFF**:
            In the image below, the target element has a raw size of ``200 x 150px``.
            Positive changes to the :attr:`spread_radius` values will cause the raw
            :attr:`size` of the shadow to increase, while negative values will cause
            the shadow to shrink.

            .. image:: images/boxshadow_spread_radius.svg
                :align: center

        | 

        - :attr:`inset` **ON**:
            Positive values will cause the shadow to grow into the bounding box,
            while negative values will cause the shadow to shrink.

            .. image:: images/boxshadow_spread_radius_inset.svg
                :align: center

        '''
        return self._spread_radius

    @spread_radius.setter
    def spread_radius(self, value):
        self._spread_radius = self._check_iter("spread_radius", value)
        self._update_shadow()

    @property
    def offset(self):
        '''Specifies shadow offsets in `[horizontal, vertical]` format. 
        Positive values for the offset indicate that the shadow should move to
        the right and/or top.
        The negative ones indicate that the shadow should move to the left
        and/or down.

        Defaults to ``(0.0, 0.0)``.

        For this property to work as expected, it is indicated that the value
        of :attr:`pos` coincides with the position of the target element of the
        shadow, as in the example below:

        - :attr:`inset` **OFF**:
            .. image:: images/boxshadow_offset.svg
                :align: center
        
        | 

        - :attr:`inset` **ON**:
            .. image:: images/boxshadow_offset_inset.svg
                :align: center

        '''
        return self._offset

    @offset.setter
    def offset(self, value):
        self._offset = self._check_iter("offset", value)
        self._update_shadow()

    @property
    def blur_radius(self):
        '''Define the shadow blur radius. Controls shadow expansion and softness.

        Defaults to ``15.0``.

        In the images below, the start and end positions of the shadow blur
        effect length are indicated.
        The transition between color and transparency is seamless, and although
        the shadow appears to end before before the dotted rectangle, its end
        is made to be as smooth as possible.

        - :attr:`inset` **OFF**:
            .. image:: images/boxshadow_blur_radius.svg
                :align: center
        
        | 

        - :attr:`inset` **ON**:
            .. image:: images/boxshadow_blur_radius_inset.svg
                :align: center
        
        | 

        .. note::
            In some cases (**if this is not your intention**), placing an element
            above the shadow (before the blur radius ends) will result in a unwanted
            cropping/overlay behavior rather than continuity, breaking the
            shadow's soft ending, as shown in the image below.

            | 

            .. image:: images/boxshadow_common_mistake_1.svg
                :align: center

        '''
        return self._blur_radius

    @blur_radius.setter
    def blur_radius(self, value):
        _blur_radius = self._check_float("blur_radius", value)
        self._blur_radius = self._bounded_value(_blur_radius, min_value=0.0)
        self._update_shadow()

    @property
    def inset(self):
        """Defines whether the shadow is drawn from the inside out or from the outline to the inside of the ``BoxShadow`` instruction.

        Defaults to ``False``.

        .. note::
            | 

            Although the inset mode determines the drawing behavior of the shadow, the position of the ``BoxShadow``
            instruction in the ``canvas`` hierarchy depends on the other graphic instructions present in the
            :class:`~kivy.graphics.instructions.Canvas` instruction tree.

            | 

            In other words, if the **target** is in the ``canvas`` layer and you want to use the default ``inset = False``
            mode to create an elevation effect, you must declare the ``BoxShadow`` instruction in ``canvas.before`` layer.

            | 

            .. image:: images/boxshadow_example_1.png
                :align: center
                :width: 300px

            .. code-block:: kv

                <MyWidget@Widget>:
                    size_hint: None, None
                    size: 100, 100
                    pos: 100, 100

                    canvas.before:
                        # BoxShadow statements
                        Color:
                            rgba: 0, 0, 0, 0.65
                        BoxShadow:
                            pos: self.pos
                            size: self.size
                            offset: 0, -10
                            blur_radius: 25
                            spread_radius: -10, -10
                            border_radius: 10, 10, 10, 10
                    
                    canvas:
                        # target element statements
                        Color:
                            rgba: 1, 1, 1, 1
                        Rectangle:
                            pos: self.pos
                            size: self.size
            
            | 

            Or, if the target is in the ``canvas`` layer and you want to use the ``inset = True`` mode to create an
            insertion effect, you must declare the ``BoxShadow`` instruction in the ``canvas`` layer, immediately after
            the **target** ``canvas`` declaration, or declare it in ``canvas.after``.

            | 

            .. image:: images/boxshadow_example_2.png
                :align: center
                :width: 300px

            .. code-block:: kv

                <MyWidget@Widget>:
                    size_hint: None, None
                    size: 100, 100
                    pos: 100, 100

                    canvas:
                        # target element statements
                        Color:
                            rgba: 1, 1, 1, 1
                        Rectangle:
                            pos: self.pos
                            size: self.size

                        # BoxShadow statements
                        Color:
                            rgba: 0, 0, 0, 0.65
                        BoxShadow:
                            inset: True
                            pos: self.pos
                            size: self.size
                            offset: 0, -10
                            blur_radius: 25
                            spread_radius: -10, -10
                            border_radius: 10, 10, 10, 10

            | 

            **In summary:**

                - Elevation effect - ``inset = False``: the ``BoxShadow`` instruction needs to be drawn **before** the target element.

                - Insertion effect - ``inset = True``: the ``BoxShadow`` instruction needs to be drawn **after** the target element.

            | 

            In general, ``BoxShadow`` is more flexible than box-shadow (CSS) because the ``inset = False`` and
            ``inset = True`` modes do not limit the drawing of the shadow below and above the target element,
            respectively. Actually, you can define any hierarchy you want in the :class:`~kivy.graphics.instructions.Canvas`
            declaration tree, to create more complex effects that go beyond common shadow effects.

        **Modes:**

            - ``False`` (default) - The shadow is drawn inside out the ``BoxShadow`` instruction, creating a raised effect.
            
            - ``True`` - The shadow is drawn from the outline to the inside of the ``BoxShadow`` instruction, creating a inset effect.

        .. image:: images/boxshadow_inset.svg
            :align: center

        """
        return self._inset

    @inset.setter
    def inset(self, value):
        self._inset = self._check_bool(value)
        self._update_shadow()



'''
Text layout
============

An internal module for laying out text according to options and constraints.
This is not part of the API and may change at any time.
'''


__all__ = ('layout_text', 'LayoutWord', 'LayoutLine')


cdef inline int max(int a, int b): return b if a <= b else a
cdef inline int min(int a, int b): return a if a <= b else b


cdef class LayoutWord:
    '''Formally describes a word contained in a line. The name word simply
    means a chunk of text and can be used to describe any text.

    A word has some width, height and is rendered according to options saved
    in :attr:`options`. See :class:`LayoutLine` for its usage.

    :Parameters:
        `options`: dict
            the label options dictionary for this word.
        `lw`: int
            the width of the text in pixels.
        `lh`: int
            the height of the text in pixels.
        `text`: string
            the text of the word.
    '''

    def __cinit__(self, dict options, int lw, int lh, object text):
        self.text = text
        self.lw = lw
        self.lh = lh
        self.options = options


cdef class LayoutLine:
    ''' Formally describes a line of text. A line of text is composed of many
    :class:`LayoutWord` instances, each with it's own text, size and options.

    A :class:`LayoutLine` instance does not always imply that the words
    contained in the line ended with a newline. That is only the case if
    :attr:`is_last_line` is True. For example a single real line of text can
    be split across multiple :class:`LayoutLine` instances if the whole line
    doesn't fit in the constrained width.

    :Parameters:
        `x`: int
            the location in a texture from where the left side of this line is
            began drawn.
        `y`: int
            the location in a texture from where the bottom of this line is
            drawn.
        `w`: int
            the width of the line. This is the sum of the individual widths
            of its :class:`LayoutWord` instances. Does not include any padding.
        `h`: int
            the height of the line. This is the maximum of the individual
            heights of its :class:`LayoutWord` instances multiplied by the
            `line_height` of these instance. So this is larger then the word
            height.
        `is_last_line`: bool
            whether this line was the last line in a paragraph. When True, it
            implies that the line was followed by a newline. Newlines should
            not be included in the text of words, but is implicit by setting
            this to True.
        `line_wrap`: bool
            whether this line is continued from a previous line which didn't
            fit into a constrained width and was therefore split across
            multiple :class:`LayoutLine` instances. `line_wrap` can be True
            or False independently of `is_last_line`.
        `words`: python list
            a list that contains only :class:`LayoutWord` instances describing
            the text of the line.

    '''

    def __cinit__(self, int x=0, int y=0, int w=0, int h=0, int is_last_line=0,
                  int line_wrap=0, list words=None):
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        self.is_last_line = is_last_line
        self.line_wrap = line_wrap
        if words is None:
            words = []
        self.words = words


cdef inline LayoutLine add_line(object text, int lw, int lh, LayoutLine line,
        list lines, dict options, float line_height, int xpad, int *w, int *h,
        int pos, int strip):
        ''' Adds to the current line the text if lw is not zero. Increases that
        line's w/h by required amount, increases global h/w by required amount
        and returns new empty line.

        pos being -1 indicates we just append line, else, we insert the line at
        index pos in lines. E.g. if we add lines from bottom up.

        This assumes that global h is accurate and includes the text previously
        added to the line.
        '''
        cdef int old_lh = line.h, count = <int>len(lines), add_h
        if lw:
            line.words.append(LayoutWord(options, lw, lh, text))
            line.w += lw

        line.h = max(int(lh * line_height), line.h)
        if count:
            add_h = line.h
        else:
            add_h = max(lh, line.h)
        # if we're appending to existing line don't add height twice
        h[0] = h[0] + add_h - old_lh
        w[0] = max(w[0], line.w + xpad)
        if strip:
            final_strip(line)
        if pos == -1:
            lines.append(line)
        else:
            lines.insert(pos, line)
        return LayoutLine()


cdef inline void final_strip(LayoutLine line):
    ''' Ensures that the line does not end with trailing spaces.

    Given the line, it'll start from the last word and strip from the
    right. If the word becomes empty, it'll remove it and strip the word
    previous to that and so on.
    '''
    cdef int diff
    cdef LayoutWord last_word
    cdef object stripped

    # XXX: here we strip any trailing spaces reducing the width of the line
    # however, the height is not reduced, even if the part that might be larger
    # is removed, potentially reducing the height of the line. It is not likely
    # a issue, but can be 'fixed' at the cost of re-computing line height

    while (len(line.words) and (line.words[-1].text.endswith(' ') or
                                line.words[-1].text == '')):
        last_word = line.words.pop()

        if last_word.text == '':  # empty str, pop it
            line.w -= last_word.lw  # likely 0
            continue

        stripped = last_word.text.rstrip()  # ends with space
        # subtract ending space length
        diff = ((<int>len(last_word.text) - <int>len(stripped)) *
                last_word.options['space_width'])
        line.w = max(0, line.w - diff)  # line w
        line.words.append(LayoutWord(   # re-add last word
        last_word.options, max(0, last_word.lw - diff),
        last_word.lh, stripped))


cdef inline layout_text_unrestricted(object text, list lines, int w, int h,
    int uh, dict options, object get_extents, int dwn, int complete,
    int xpad, int max_lines, float line_height, int strip):
    ''' Layout when the width is unrestricted; text_size[0] is None.
    It's a bit faster.
    '''

    cdef list new_lines
    cdef int s, lw, lh, old_lh, i = -1, n
    cdef int lhh, k, pos, add_h
    cdef object line, val = False, indices
    cdef LayoutLine _line

    new_lines = text.split('\n')
    n = <int>len(new_lines)
    s = 0
    k = n
    pos = <int>len(lines)  # always include first line, start w/ no lines added
    # there's a last line to which first (last) new line must be appended
    if pos:
        if dwn:  # append to last line
            _line = lines[-1]
            line = new_lines[0]
            s = 1
        else:  # append to first line
            _line = lines[0]
            line = new_lines[-1]
            k = n - 1
        if strip:
            if not _line.w:  # no proceeding text: strip leading
                line = line.lstrip()
            # ends this line so right strip
            if complete or (dwn and n > 1 or not dwn and pos > 1):
                line = line.rstrip()
        lw, lh = get_extents(line)

        old_lh = _line.h
        if lw:  # when adding to existing line, don't check uh
            _line.words.append(LayoutWord(options, lw, lh, line))
            _line.w += lw
            _line.h = max(int(lh * line_height), _line.h)
            if pos == 1:  # still first line
                add_h = max(lh, _line.h)
            else:
                add_h = _line.h
        elif strip and (complete or (dwn and n > 1 or not dwn and pos > 1)):
            # if we finish this line, make sure it doesn't end in spaces
            final_strip(_line)
            add_h = _line.h
        else:
            add_h = _line.h

        w = max(w, _line.w + xpad)
        h += add_h - old_lh

    # now do the remaining lines
    indices = range(s, k) if dwn else reversed(range(s, k))
    for i in indices:
        # always compute first line, even if it won't be displayed
        if (max_lines > 0 and pos + 1 > max_lines or pos and
            uh != -1 and h > uh):
            i += -1 if dwn else 1
            break
        line = new_lines[i]
        # the last line is only stripped from left
        if strip:
            if complete or (dwn and i < n - 1 or not dwn and i > s):
                line = line.strip()
            else:
                line = line.lstrip()
        lw, lh = get_extents(line)

        lhh = int(lh * line_height)
        if pos:
            add_h = lhh
        else:  # for the first line, always use full height
            add_h = lh
        if uh != -1 and h + add_h > uh and pos:  # too high
            i += -1 if dwn else 1
            break

        pos += 1
        w = max(w, int(lw + xpad))
        h += add_h
        if lw:
            _line = LayoutLine(0, 0, lw, lhh, 1, 0, [LayoutWord(options, lw,
                                                                lh, line)])
        else:
            _line = LayoutLine(0, 0, 0, lhh, 1, 0, [])
        new_lines[i] = _line

    if s != k:
        if dwn:
            lines.extend(new_lines[s:i + 1])
            val = i != k - 1
        else:
            if k != i:
                lines.extend([None, ] * (k - i))
                lines[(k - i):] = lines[:len(lines) - (k - i)]
                lines[:(k - i)] = new_lines[i:k]
            val = i != 0

    return w, h, val


def layout_text(object text, list lines, tuple size, tuple text_size,
                dict options, object get_extents, int append_down,
                int complete):
    ''' Lays out text into a series of :class:`LayoutWord` and
    :class:`LayoutLine` instances according to the options specified.

    The function is designed to be called many times, each time new text is
    appended to the last line (or first line if appending upwards), unless a
    newline is present in the text. Each text appended is described by its own
    options which can change between successive calls. If the text is
    constrained, we stop as soon as the constraint is reached.

    :Parameters:
        `text`: string or bytes
            the text to be broken down into lines. If lines is not empty,
            the text is added to the last line (or first line if `append_down`
            is False) until a newline is reached which creates a new line in
            `lines`. See :class:`LayoutLine`.
        `lines`: list
            a list of :class:`LayoutLine` instances, each describing a line of
            the text. Calls to :func:`layout_text` append or create
            new :class:`LayoutLine` instances in `lines`.
        `size`: 2-tuple of ints
            the size of the laid out text so far. Upon first call it should
            probably be (0, 0), afterwards it should be the (w, h) returned
            by this function in a previous call. When size reaches the
            constraining size, `text_size`, we stop adding lines and return
            True for the clipped parameter. size includes the x and y padding.
        `text_size`: 2-tuple of ints or None.
            the size constraint on the laid out text. If either element is
            None, the text is not constrained in that dimension. For example,
            (None, 200) will constrain the height, including padding to 200,
            while the width is unconstrained. The first line, and the first
            character of a line is always returned, even if it exceeds the
            constraint. The value be changed between different calls.
        `options`: dict
            the label options of this `text`. The options are saved with each
            word allowing different words to have different options from
            successive calls.

            Note, `options` must include a `space_width` key with a value
            indicating the width of a space for that set of options.
        `get_extents`: callable
            a function called with a string, which returns a tuple containing
            the width, height of the string.
        `append_down`: bool
            Whether successive calls to the function appends lines before or
            after the existing lines. If True, they are appended to the last
            line and below it. If False, it's appended at the first line and
            above. For example, if False, everything after the last
            newline in `text` is appended to the first line in lines.
            Everything before the last newline is inserted at the start of
            lines in same order as text; that is we do not invert the line
            order.

            This allows laying out from top to bottom until the constrained is
            reached, or from bottom to top until the constrained is reached.
        `complete`: bool
            whether this text complete lines. It use is that normally is
            strip in `options` is True, all leading and trailing spaces
            are removed from each line except from the last line (or first
            line if `append_down` is False) which only removes leading spaces.
            That's because further text can still be appended to the last line
            so we cannot strip them. If `complete` is True, it indicates no
            further text is coming and all lines will be stripped.

            The function can also be called with `text` set to the empty string
            and `complete` set to True in order for the last (first) line to
            be stripped.

    :returns:
        3-tuple, (w, h, clipped).
        w and h is the width and height of the text in lines so far and
        includes padding. This can be larger than `text_size`, e.g. if not even
        a single fitted, the first line would still be returned.
        `clipped` is True if not all the text has been added to lines because
        w, h reached the constrained size.

    Following is a simple example with no padding and no stripping::

        >>> from kivy.core.text import Label
        >>> from kivy.core.text.text_layout import layout_text

        >>> l = Label()
        >>> lines = []
        >>> # layout text with width constraint by 50, but no height constraint
        >>> w, h, clipped = layout_text('heres some text\\nah, another line',
        ... lines, (0, 0), (50, None), l.options, l.get_cached_extents(), True,
        ... False)
        >>> w, h, clipped
        (46, 90, False)
        # now add text from bottom up, and constrain width only be 100
        >>> w, h, clipped = layout_text('\\nyay, more text\\n', lines, (w, h),
        ... (100, None), l.options, l.get_cached_extents(), False, True)
        >>> w, h, clipped
        (77, 120, 0)
        >>> for line in lines:
        ...     print('line w: {}, line h: {}'.format(line.w, line.h))
        ...     for word in line.words:
        ...         print('w: {}, h: {}, text: {}'.format(word.lw, word.lh,
        ...         [word.text]))
        line w: 0, line h: 15
        line w: 77, line h: 15
        w: 77, h: 15, text: ['yay, more text']
        line w: 31, line h: 15
        w: 31, h: 15, text: ['heres']
        line w: 34, line h: 15
        w: 34, h: 15, text: [' some']
        line w: 24, line h: 15
        w: 24, h: 15, text: [' text']
        line w: 17, line h: 15
        w: 17, h: 15, text: ['ah,']
        line w: 46, line h: 15
        w: 46, h: 15, text: [' another']
        line w: 23, line h: 15
        w: 23, h: 15, text: [' line']
    '''

    cdef int uw, uh,  _do_last_line, lwe, lhe, ends_line, is_last_line
    cdef int xpad = options['padding'][0] + options['padding'][2], ypad = options['padding'][1] + options['padding'][3]
    cdef int max_lines = int(options.get('max_lines', 0))
    cdef float line_height = options['line_height']
    cdef int strip = options['strip'] or options['halign'] == 'justify'
    cdef int ref_strip = options['strip_reflow']
    cdef int w = size[0], h = size[1]  # width and height of the texture so far
    cdef list new_lines
    cdef int s, lw = -1, lh = -1, old_lh, i = -1, n, m, e
    cdef int lhh, lww, k, bare_h, dwn = append_down, pos = 0
    cdef object line, ln, val, indices
    cdef LayoutLine _line
    cdef int is_space
    uw = text_size[0] if text_size[0] is not None else -1
    uh = text_size[1] if text_size[1] is not None else -1

    if not h:
        h = ypad

    if uw == -1:  # no width specified
        return layout_text_unrestricted(text, lines, w, h, uh, options,
        get_extents, dwn, complete, xpad, max_lines, line_height, strip)


    new_lines = text.split('\n')
    n = <int>len(new_lines)
    uw = max(0, uw - xpad)  # actual w, h allowed for rendering
    _, bare_h = get_extents('')
    if dwn:
        pos = -1  # don't use pos when going down b/c we append at end of lines

    # split into lines and find how many line wraps each line requires
    indices = range(n) if dwn else reversed(range(n))
    for i in indices:
        is_space = 0
        k = <int>len(lines)
        if (max_lines > 0 and k > max_lines or uh != -1 and
            h > uh and k > 1):
            break

        is_last_line = not (dwn and i < n - 1 or not dwn and i)
        # whether this line is ended, or if we may append to it later
        ends_line = complete or not is_last_line
        if not dwn:  # new line will be appended at top, unless changed below
            pos = 0
        # for the first (last if not down) new line, append it to previous line
        if (i and dwn or not dwn and i != n - 1) or not k:  # interior line
            _line = LayoutLine()
        else:
            if dwn:  # take last line
                _line = lines.pop()
            else:  # need to append right before 1st line ends in case of wrap
                while pos + 1 < k and not lines[pos + 1].line_wrap:
                    pos += 1
                _line = lines.pop(pos)

        line = new_lines[i]
        # there's no proceeding text, so strip leading
        if not _line.w and (strip or ref_strip and _line.line_wrap):
            line = line.lstrip()
        if strip and ends_line:
            line = line.rstrip()

        k = <int>len(line)
        if not k:  # just add empty line if empty
            _line.is_last_line = ends_line  # nothing will be appended
            # ensure we don't leave trailing from before
            _line = add_line('', 0, bare_h, _line, lines, options, line_height,
                             xpad, &w, &h, pos, _line.w and ends_line)
            continue

        '''----------------- we now a non-empty line ------------------------
        what we do is given the current text in this real line if we can fit
        another word, add it. Otherwise add it to a new line. But if a single
        word doesn't fit on a single line, just split the word itself into
        multiple lines'''

        # s is idx in line of start of this actual line, e is idx of
        # next space, m is idx after s that still fits on this line
        s = m = e = 0
        while s != k:
            # find next space or end, if end don't keep checking
            if e != k:
                # leading spaces
                if s == m and not _line.w and line[s] == ' ' and (strip or ref_strip and _line.line_wrap):
                    s = m = s + 1
                    # trailing spaces were stripped, so end is always not space
                    continue

                # when not stripping, if we found a space last, don't jump to
                # the next space, but instead move pos to after this space, to
                # allow fitting this space on the current line
                if strip or not is_space:
                    e = line.find(' ', m + 1)
                    is_space = 1
                else:
                    e = m + 1
                    is_space = 0
                if e is -1:
                    e = k

            lwe, lhe = get_extents(line[s:e])  # does next word fit?
            if lwe + _line.w > uw:  # too wide
                ln = ''
                lww, lhh = 0, bare_h
                _do_last_line = 0
                # if there's already some text, commit and go next line
                if s != m:
                    _do_last_line = 1
                    if (strip or ref_strip) and line[m - 1] == ' ':
                        ln = line[s:m].rstrip()
                        lww, lhh = get_extents(ln)
                    else:
                        ln = line[s:m]
                        lww, lhh = lw, lh
                    s = m
                elif _line.w:
                    _do_last_line = 1

                if _do_last_line:
                    # if there's proceeding text and ln is '': strip trailing
                    _line = add_line(ln, lww, lhh, _line, lines, options,
                        line_height, xpad, &w, &h, pos, _line.w and not lww)
                    _line.line_wrap = 1
                    if not dwn:
                        pos += 1

                # try to fit word on new line, if it doesn't fit we'll
                # have to break the word into as many lines needed
                if strip or ref_strip and _line.line_wrap:
                    s = e - <int>len(line[s:e].lstrip())
                if s == e:  # if it was only a stripped space, move on
                    m = s
                    continue

                # now break single word into as many lines needed
                m = s
                while s != e:
                    # does remainder fit in single line?
                    lwe, lhe = get_extents(line[s:e])
                    if lwe + _line.w <= uw:
                        m = e
                        break
                    # if not, fit as much as possible into this line
                    while (m != e and
                           get_extents(line[s:m + 1])[0] + _line.w <= uw):
                        m += 1
                    # not enough room for even single char, skip it
                    if m == s:
                        m += 1
                    _line.is_last_line = m == k  # is last line?
                    lww, lhh = get_extents(line[s:m])
                    _line = add_line(line[s:m], lww, lhh, _line, lines,
                        options, line_height, xpad, &w, &h, pos, 0)
                    _line.line_wrap = 1
                    if not dwn:
                        pos += 1
                    s = m
                m = s  # done with long word, go back to normal

            else:   # the word fits
                # don't allow leading spaces on empty lines
                #if strip and m == s and line[s:e] == ' ' and not _line.w:
                if (strip or ref_strip and _line.line_wrap) and line[s:e] == ' ' and not _line.w:
                    s = m = e
                    continue
                m = e

            if m == k:  # we're done
                if s != k or _line.w:
                    _line.is_last_line = ends_line  # line end
                    _line = add_line(line[s:], lwe, lhe, _line, lines, options,
                                     line_height, xpad, &w, &h, pos, 0)
                break
            lw, lh = lwe, lhe  # save current lw/lh, then fit more in line

    val = dwn and i != n - 1 or not dwn and i
    # ensure the number of lines is not more than the user asked
    # above, we might have gone a few lines over
    if max_lines > 0 and len(lines) > max_lines:
        val = True
        if dwn:
            h-= sum(l.h for l in lines[max_lines:])
            del lines[max_lines:]
        else:
            h-= sum(l.h for l in lines[:max(0, <int>len(lines) - max_lines)])
            del lines[:max(0, <int>len(lines) - max_lines)]

    # now make sure we don't have lines outside specified height
    k = <int>len(lines)
    if k > 1 and uh != -1 and h > uh:
        val = True
        if dwn:  # remove from last line going up
            i = k -1  # will removing the ith line make it fit?
            while i > 0 and h > uh:
                h -= lines[i].h
                i -= 1
            del lines[i + 1:]  # we stopped when keeping the ith line still fits
        else:  # remove from first line going down
            i = 0  # will removing the ith line make it fit?
            while i < k - 1 and h > uh:
                h -= lines[i].h
                i += 1
            del lines[:i]

    return w, h, val



'''
SDL2 audio provider
===================

This core audio implementation require SDL_mixer library.
It might conflict with any other library that are using SDL_mixer, such as
ffmpeg-android.

Native formats:

* wav, since 1.9.0

Depending the compilation of SDL2 mixer and/or installed libraries:

* ogg since 1.9.1 (mixer needs libvorbis/libogg)
* flac since 1.9.1 (mixer needs libflac)
* mp3 since 1.9.1 (mixer needs libsmpeg/libmad; only use mad for GPL apps)
  * Since 1.10.1 + mixer 2.0.2, mpg123 can also be used
* sequenced formats since 1.9.1 (midi, mod, s3m, etc. Mixer needs
  libmodplug or libmikmod)

.. Note::

    Sequenced format support changed with mixer v2.0.2. If mixer is
    linked with one of libmodplug or libmikmod, format support for
    both libraries is assumed. This will work perfectly with formats
    supported by both libraries, but if you were to try to load an
    obscure format (like `apun` file with mikmod only), it will fail.

    * Kivy <= 1.10.0: will fail to build with mixer >= 2.0.2
      will report correct format support with < 2.0.2
    * Kivy >= 1.10.1: will build with old and new mixer, and
      will "guesstimate" sequenced format support

.. Warning::

    Sequenced formats use the SDL2 Mixer music channel, you can only play
    one at a time, and .length will be -1 if music fails to load, and 0
    if loaded successfully (we can't get duration of these formats)
'''

__all__ = ('SoundSDL2', 'MusicSDL2')

include "../../../kivy/lib/sdl2.pxi"
include "../../../kivy/graphics/common.pxi"  # For malloc and memcpy (on_pitch)

from kivy.core.audio import Sound, SoundLoader
from kivy.logger import Logger
from kivy.clock import Clock

cdef int mix_is_init = 0
cdef int mix_flags = 0


cdef mix_init():
    cdef int audio_rate = 44100
    cdef unsigned short audio_format = AUDIO_S16SYS
    cdef int audio_channels = 2
    cdef int audio_buffers = 4096
    cdef int want_flags = 0
    global mix_is_init
    global mix_flags

    # avoid next call
    if mix_is_init != 0:
        return

    if SDL_Init(SDL_INIT_AUDIO) < 0:
        Logger.critical('AudioSDL2: Unable to initialize SDL: {}'.format(
                        SDL_GetError()))
        mix_is_init = -1
        return 0

    # In mixer 2.0.2, MIX_INIT_MODPLUG is now implied by MIX_INIT_MOD,
    # and MIX_INIT_FLUIDSYNTH was renamed to MIX_INIT_MID. In previous
    # versions, we requested both _MODPLUG and _MOD + _FLUIDSYNTH.
    # 0x20 used to be MIX_INIT_FLUIDSYNTH, now MIX_INIT_MID
    # 0x4  used to be MIX_INIT_MODPLUG before 2.0.2
    want_flags = MIX_INIT_FLAC | MIX_INIT_OGG | MIX_INIT_MP3
    want_flags |= MIX_INIT_MOD | 0x20 | 0x4

    mix_flags = Mix_Init(want_flags)

    if Mix_OpenAudio(audio_rate, audio_format, audio_channels, audio_buffers):
        Logger.critical('AudioSDL2: Unable to open mixer: {}'.format(
                        Mix_GetError()))
        mix_is_init = -1
        return 0

    mix_is_init = 1
    return 1

# Container for samples (Mix_LoadWAV)
cdef class ChunkContainer:
    cdef Mix_Chunk *chunk
    cdef Mix_Chunk *original_chunk
    cdef int channel

    def __init__(self):
        self.chunk = NULL
        self.channel = -1

    def __dealloc__(self):
        if self.chunk != NULL:
            if Mix_GetChunk(self.channel) == self.chunk:
                Mix_HaltChannel(self.channel)
            Mix_FreeChunk(self.chunk)
            self.chunk = NULL
        if self.original_chunk != NULL:
            Mix_FreeChunk(self.original_chunk)
            self.original_chunk = NULL

# Container for music (Mix_LoadMUS), one channel only
cdef class MusicContainer:
    cdef Mix_Music *music
    cdef int playing

    def __init__(self):
        self.music = NULL
        self.playing = 0

    def __dealloc__(self):
        if self.music != NULL:
            # I think FreeMusic halts automatically, probably not needed
            if Mix_PlayingMusic() and self.playing:
                Mix_HaltMusic()
            Mix_FreeMusic(self.music)
            self.music = NULL


class SoundSDL2(Sound):

    @staticmethod
    def extensions():
        mix_init()
        extensions = ["wav"]
        if mix_flags & MIX_INIT_FLAC:
            extensions.append("flac")
        if mix_flags & MIX_INIT_MP3:
            extensions.append("mp3")
        if mix_flags & MIX_INIT_OGG:
            extensions.append("ogg")
        return extensions

    def __init__(self, **kwargs):
        self._check_play_ev = None
        self.cc = ChunkContainer()
        mix_init()
        super(SoundSDL2, self).__init__(**kwargs)

    def _check_play(self, dt):
        cdef ChunkContainer cc = self.cc
        if cc.channel == -1 or cc.chunk == NULL:
            return False
        if Mix_Playing(cc.channel):
            return
        if self.loop:
            def do_loop(dt):
                self.play()
            Clock.schedule_once(do_loop)
        else:
            self.stop()
        return False

    def _get_length(self):
        cdef ChunkContainer cc = self.cc
        cdef int freq, channels
        cdef unsigned int points, frames
        cdef unsigned short fmt
        if cc.chunk == NULL:
            return 0
        if not Mix_QuerySpec(&freq, &fmt, &channels):
            return 0
        points = <unsigned int>int(cc.chunk.alen / ((fmt & 0xFF) / 8))
        frames = <unsigned int>int(points / channels)
        return <double>frames / <double>freq

    def on_pitch(self, instance, value):
        cdef ChunkContainer cc = self.cc
        cdef int freq, channels
        cdef unsigned short fmt
        cdef SDL_AudioCVT cvt
        if cc.chunk == NULL:
            return
        if not Mix_QuerySpec(&freq, &fmt, &channels):
            return
        SDL_BuildAudioCVT(
            &cvt,
            fmt, channels, int(freq * self.pitch),
            fmt, channels, freq,
        )
        cvt.buf = <Uint8 *>malloc(cc.original_chunk.alen * cvt.len_mult)
        cvt.len = cc.original_chunk.alen
        memcpy(cvt.buf, cc.original_chunk.abuf, cc.original_chunk.alen)
        SDL_ConvertAudio(&cvt)
        cc.chunk = Mix_QuickLoad_RAW(cvt.buf, <Uint32>(cvt.len * cvt.len_ratio))

    def play(self):
        cdef ChunkContainer cc = self.cc
        self.stop()
        if cc.chunk == NULL:
            return
        cc.chunk.volume = int(self.volume * 128)
        cc.channel = Mix_PlayChannel(-1, cc.chunk, 0)
        if cc.channel == -1:
            Logger.warning('AudioSDL2: Unable to play {}: {}'.format(
                           self.source, Mix_GetError()))
            return
        # schedule event to check if the sound is still playing or not
        self._check_play_ev = Clock.schedule_interval(self._check_play, 0.1)
        super(SoundSDL2, self).play()

    def stop(self):
        cdef ChunkContainer cc = self.cc
        if cc.chunk == NULL or cc.channel == -1:
            return
        if Mix_GetChunk(cc.channel) == cc.chunk:
            Mix_HaltChannel(cc.channel)
        cc.channel = -1
        if self._check_play_ev is not None:
            self._check_play_ev.cancel()
            self._check_play_ev = None
        super(SoundSDL2, self).stop()

    def load(self):
        cdef ChunkContainer cc = self.cc
        self.unload()
        if self.source is None:
            return

        if isinstance(self.source, bytes):
            fn = self.source
        else:
            fn = self.source.encode('UTF-8')

        cc.chunk = Mix_LoadWAV(<char *><bytes>fn)
        if cc.chunk == NULL:
            Logger.warning('AudioSDL2: Unable to load {}: {}'.format(
                           self.source, Mix_GetError()))
        else:
            cc.original_chunk = Mix_QuickLoad_RAW(cc.chunk.abuf, cc.chunk.alen)
            cc.chunk.volume = int(self.volume * 128)
            if self.pitch != 1.:
                self.on_pitch(self, self.pitch)

    def unload(self):
        cdef ChunkContainer cc = self.cc
        self.stop()
        if cc.chunk != NULL:
            Mix_FreeChunk(cc.chunk)
            cc.chunk = NULL

    def on_volume(self, instance, volume):
        cdef ChunkContainer cc = self.cc
        if cc.chunk != NULL:
            cc.chunk.volume = int(volume * 128)


# LoadMUS supports OGG, MP3, WAV but we only use it for native midi,
# libmikmod, libmodplug and libfluidsynth to avoid confusion
class MusicSDL2(Sound):

    @staticmethod
    def extensions():
        mix_init()
        # FIXME: this should probably evolve to use the new has_music()
        #        interface to determine format support

        # Assume native midi support (defaults to enabled), but may use
        # modplug, fluidsynth or timidity in reality. It may also be
        # disabled completely, in which case loading it will fail
        extensions = set(['mid', 'midi'])

        # libmodplug and libmikmod, may be incomplete.
        # 0x4 is for mixer < 2.0.2, MIX_INIT_MODPLUG
        if mix_flags & (MIX_INIT_MOD | 0x4):
            extensions.update(['669', 'abc', 'amf', 'ams', 'apun', 'dbm',
                               'dmf', 'dsm', 'far', 'gdm', 'it',   'j2b',
                               'mdl', 'med', 'mod', 'mt2', 'mtm',  'okt',
                               'pat', 'psm', 'ptm', 's3m', 'stm',  'stx',
                               'ult', 'umx', 'uni', 'xm'])

        return list(extensions)

    def __init__(self, **kwargs):
        self.mc = MusicContainer()
        self._check_play_ev = None
        mix_init()
        super(MusicSDL2, self).__init__(**kwargs)

    def _check_play(self, dt):
        cdef MusicContainer mc = self.mc
        if mc.music == NULL:
            return False
        if mc.playing and Mix_PlayingMusic():
            return
        if self.loop:
            def do_loop(dt):
                self.play()
            Clock.schedule_once(do_loop)
        else:
            self.stop()
        return False

    # No way to check length; return -1 if music is loaded, 0 otherwise
    def _get_length(self):
        cdef MusicContainer mc = self.mc
        if mc.music == NULL:
            return -1
        return 0

    def play(self):
        cdef MusicContainer mc = self.mc
        self.stop()
        if mc.music == NULL:
            return
        Mix_VolumeMusic(int(self.volume * 128))
        if Mix_PlayMusic(mc.music, 1) == -1:
            Logger.warning('AudioSDL2: Unable to play music {}: {}'.format(
                           self.source, Mix_GetError()))
            return
        mc.playing = 1
        # schedule event to check if the sound is still playing or not
        self._check_play_ev = Clock.schedule_interval(self._check_play, 0.1)
        super(MusicSDL2, self).play()

    def stop(self):
        cdef MusicContainer mc = self.mc
        if mc.music == NULL or not mc.playing:
            return
        Mix_HaltMusic()
        mc.playing = 0
        if self._check_play_ev is not None:
            self._check_play_ev.cancel()
            self._check_play_ev = None
        super(MusicSDL2, self).stop()

    def load(self):
        cdef MusicContainer mc = self.mc
        self.unload()
        if self.source is None:
            return

        if isinstance(self.source, bytes):
            fn = self.source
        else:
            fn = self.source.encode('UTF-8')

        mc.music = Mix_LoadMUS(<char *><bytes>fn)
        if mc.music == NULL:
            Logger.warning('AudioSDL2: Unable to load music {}: {}'.format(
                           self.source, Mix_GetError()))
        else:
            Mix_VolumeMusic(int(self.volume * 128))

    def unload(self):
        cdef MusicContainer mc = self.mc
        self.stop()
        if mc.music != NULL:
            Mix_FreeMusic(mc.music)
            mc.music = NULL

    def on_volume(self, instance, volume):
        cdef MusicContainer mc = self.mc
        if mc.music != NULL and mc.playing:
            Mix_VolumeMusic(int(volume * 128))


SoundLoader.register(SoundSDL2)
SoundLoader.register(MusicSDL2)





__all__ = ('ClockNotRunningError', 'ClockEvent', 'CyClockBase',
           'FreeClockEvent', 'CyClockBaseFree')


cdef extern from "float.h":
    double DBL_MAX

from kivy.weakmethod import WeakMethod
from kivy.logger import Logger
from threading import Lock, RLock


class ClockNotRunningError(RuntimeError):
    """Raised by the kivy Clock when scheduling an event if the
    Kivy Clock has already finished (:class:`~CyClockBase.stop_clock` was
    called).
    """


cdef class ClockEvent(object):
    ''' A class that describes a callback scheduled with kivy's :attr:`Clock`.
    This class is never created by the user; instead, kivy creates and returns
    an instance of this class when scheduling a callback.

    An event can be triggered (scheduled) by calling it. If it's already
    scheduled, nothing will happen, otherwise it'll be scheduled. E.g.::

        event = Clock.schedule_once(my_callback, .5)
        event()  # nothing will happen since it's already scheduled.
        event.cancel()  # cancel it
        event()  # now it's scheduled again.
    '''

    def __init__(
            self, CyClockBase clock, int loop, callback, double timeout,
            double starttime, cid=None, int trigger=False,
            clock_ended_callback=None, release_ref=True, **kwargs):
        super(ClockEvent, self).__init__(**kwargs)
        self._is_triggered = False
        self.next = None
        self.prev = None
        self.cid = None
        self.clock = clock
        self.loop = loop
        self.weak_callback = None
        self.callback = callback
        self.timeout = timeout
        self._last_dt = starttime
        self._dt = 0.
        self.weak_clock_ended_callback = None
        self.clock_ended_callback = clock_ended_callback
        self.release_ref = release_ref

        if trigger:
            clock._lock_acquire()
            # only fail if clock_ended_callback was provided from the lifecycle_aware API
            if self.clock.has_ended and (
                    self.clock_ended_callback is not None or
                    self.weak_clock_ended_callback is not None):
                clock._lock_release()
                raise ClockNotRunningError

            self._is_triggered = True

            if clock._root_event is None:
                clock._last_event = clock._root_event = self
            else:
                clock._last_event.next = self
                self.prev = clock._last_event
                clock._last_event = self
            clock._lock_release()

            self.clock.on_schedule(self)

    def __call__(self, *largs):
        ''' Schedules the callback associated with this instance.
        If the callback is already scheduled, it will not be scheduled again.
        '''
        # The lock must be held continuously between checking whether the clock
        # ended and triggering the event. Otherwise, if the clock has not ended,
        # then in between checking whether it ended and scheduling the event in
        # this thread, the main thread could have stopped the clock and
        # processed all the clock_ended_callbacks. So this event would be scheduled,
        # but callback and clock_ended_callback won't be called.
        self.clock._lock_acquire()
        # only fail if clock_ended_callback was provided from the lifecycle_aware API
        if self.clock.has_ended and (
                self.clock_ended_callback is not None or
                self.weak_clock_ended_callback is not None):
            self.clock._lock_release()
            raise ClockNotRunningError

        trigger = not self._is_triggered
        if trigger:
            self._is_triggered = True
            self._last_dt = self.clock._last_tick

            if self.clock._root_event is None:
                self.clock._last_event = self.clock._root_event = self
            else:
                self.clock._last_event.next = self
                self.prev = self.clock._last_event
                self.clock._last_event = self

        self.clock._lock_release()

        if trigger:
            self.clock.on_schedule(self)

    cpdef get_callback(self):
        '''Returns the callback associated with the event. Callbacks get stored
        with a indirect ref so that it doesn't keep objects alive. If the callback
        is dead, None is returned.
        '''
        cdef object callback = self.callback
        if callback is not None:
            return callback
        return self.weak_callback()

    cpdef get_clock_ended_callback(self):
        """Returns the clock_ended_callback associated with the event.
        Callbacks get stored with a indirect ref so that it doesn't keep
        objects alive. If the callback is dead or wasn't provided,
        None is returned.
        """
        cdef object callback = self.clock_ended_callback
        if callback is not None:
            return callback
        if self.weak_clock_ended_callback is None:
            return None
        return self.weak_clock_ended_callback()

    @property
    def is_triggered(self):
        '''Returns whether the event is scheduled to have its callback executed by
        the kivy thread.
        '''
        return self._is_triggered

    cpdef cancel(self):
        ''' Cancels the callback if it was scheduled to be called. If not
        scheduled, nothing happens.
        '''
        self.clock._lock_acquire()
        if self._is_triggered:
            self._is_triggered = False

            # update the cap of the list
            if self.clock._cap_event is self:
                # cap is next, so we have reached the end of the list
                # because current one being processed is going to be the last event now
                if self.clock._cap_event is self.clock._next_event:
                    self.clock._cap_event = None
                else:
                    self.clock._cap_event = self.prev  # new cap

            # update the next event pointer
            if self.clock._next_event is self:
                self.clock._next_event = self.next

            if self.prev is None:  # we're first
                if self.next is None:  # we're also last
                    self.clock._last_event = self.clock._root_event = None
                else:  # there are more past us
                    self.clock._root_event = self.next
                    self.next.prev = None
            else:  # there are some behind us
                if self.next is None:  # we are last
                    self.clock._last_event = self.prev
                    self.prev.next = None
                else:  # we are in middle
                    self.prev.next = self.next
                    self.next.prev = self.prev
            self.prev = self.next = None

        self.clock._lock_release()

    cpdef release(self):
        '''(internal method) Converts the callback into a indirect ref.
        '''
        self.weak_callback = WeakMethod(self.callback)
        self.callback = None

        if self.clock_ended_callback is not None:
            self.weak_clock_ended_callback = WeakMethod(self.clock_ended_callback)
            self.clock_ended_callback = None

    cpdef tick(self, double curtime):
        '''(internal method) Processes the event for the kivy thread.
        '''
        cdef object callback, ret
        # timeout happened ? if less than resolution process it
        if curtime - self._last_dt < self.timeout - self.clock.get_resolution():
            return True

        # calculate current timediff for this event
        self._dt = curtime - self._last_dt
        self._last_dt = curtime

        # get the callback
        callback = self.get_callback()
        if callback is None:
            self.cancel()
            return self.loop

        # if it's a trigger, allow to retrigger inside the callback
        # we have to remove event here, otherwise, if we remove later, the user
        # might have canceled in the callback and then re-triggered. That'd
        # result in the removal of the re-trigger
        if not self.loop:
            self.cancel()

        # call the callback
        # this may raise an exception, but the state must remain clean so we
        # can still cancel the event if the exception is handled and suppressed
        ret = callback(self._dt)

        # if the user returns False explicitly, remove the event
        if self.loop and ret is False:
            self.cancel()
            return False
        return self.loop

    def __repr__(self):
        return '<ClockEvent ({}) callback={}>'.format(self.timeout, self.get_callback())


cdef class FreeClockEvent(ClockEvent):
    '''The event returned by the ``Clock.XXX_free`` methods of
    :class:`CyClockBaseFree`. It stores whether the event was scheduled as a
    free event.
    '''

    def __init__(self, free, *largs, **kwargs):
        self.free = free
        ClockEvent.__init__(self, *largs, **kwargs)


cdef class CyClockBase(object):
    '''The base clock object with event support.
    '''

    def __cinit__(self, **kwargs):
        self.clock_resolution = -1
        self._max_fps = 60
        self.max_iteration = 20

    def __init__(self, **kwargs):
        super(CyClockBase, self).__init__(**kwargs)
        self._root_event = None
        self._last_event = None
        self._next_event = None
        self._cap_event = None
        self._lock = Lock()
        self._lock_acquire = self._lock.acquire
        self._lock_release = self._lock.release
        self._del_queue = []
        self._del_safe_lock = RLock()
        self._del_safe_done = False
        self.has_ended = False
        self.has_started = False

    def start_clock(self):
        """Must be called to start the clock.

        Once :meth:`stop_clock` is called, it cannot be started again.
        """
        self._lock_acquire()
        try:
            if self.has_started:
                return

            self.has_started = True
            # for now don't raise an exception when restarting because kivy's
            # graphical tests are not setup to handle clock isolation so they try
            # to restart the clock
            # if self.has_ended:
            #     raise TypeError('Clock already ended. Cannot re-start the Clock')
        finally:
            self._lock_release()

    def stop_clock(self):
        """Stops the clock and cleans up.

        This must be called to process the lifecycle_aware callbacks etc.
        """
        self._lock_acquire()
        try:
            if self.has_ended:
                return

            self.has_ended = True
        finally:
            self._lock_release()
        self._process_clock_ended_callbacks()

    cpdef get_resolution(self):
        '''Returns the minimum resolution the clock has. It's a function of
        :attr:`clock_resolution` and ``maxfps`` provided at the config.
        '''
        cdef double resolution = self.clock_resolution
        # timeout happened ? (check also if we would miss from 5ms) this
        # 5ms increase the accuracy if the timing of animation for
        # example.
        if resolution < 0:
            if self._max_fps:
                resolution = 1 / (3. * self._max_fps)
            else:
                resolution = 0.0001
        return resolution

    def on_schedule(self, event):
        '''Function that is called internally every time an event is triggered
        for this clock. It takes the event as a parameter.

        The order of ``on_schedule`` calls are not guaranteed to be in the same
        order that the events are scheduled. Similarly, it is possible that the
        event being scheduled was canceled before this is called on the event.
        That's because :meth:`on_schedule` may be called from different threads.
        '''
        pass

    cpdef ClockEvent create_lifecycle_aware_trigger(
            self, callback, clock_ended_callback, timeout=0, interval=False,
            release_ref=True):
        '''Create a Trigger event similarly to :meth:`create_trigger`, but the event
        is sensitive to the clock's state.

        If this event is triggered after the clock has stopped (:meth:`stop_clock`), then a
        :class:`ClockNotRunningError` will be raised. If the error is not raised,
        then either ``callback`` or ``clock_ended_callback`` will be
        called. ``callback`` will be called when the event
        is normally executed. If the clock is stopped before it can be executed,
        provided the app exited normally without crashing and the event wasn't manually
        canceled, and the callbacks are not garbage collected then
        ``clock_ended_callback`` will be called instead when the clock is stopped.

        :Parameters:

            `callback`: callable
                The callback to execute from kivy. It takes a single parameter - the
                current elapsed kivy time.
            `clock_ended_callback`: callable
                A callback that will be called if the clock is stopped
                while the event is still scheduled to be called. The callback takes
                a single parameter - the event object. When the event is successfully
                scheduled, if the app exited normally and the event wasn't canceled,
                and the callbacks are not garbage collected - it is guaranteed that
                either ``callback`` or ``clock_ended_callback`` would have been called.
            `timeout`: float
                How long to wait before calling the callback.
            `interval`: bool
                Whether the callback should be called once (False) or repeatedly
                with a period of ``timeout`` (True) like :meth:`schedule_interval`.
            `release_ref`: bool
                If True, the default, then if ``callback`` or ``clock_ended_callback``
                is a class method and the object has no references to it, then
                the object may be garbage collected and the callbacks won't be called.
                If False, the clock keeps a reference to the object preventing it
                from being garbage collected - so it will be called.

        :returns:

            A :class:`ClockEvent` instance. To schedule the callback of this
            instance, you can call it.

        .. versionadded:: 2.0.0
        '''
        cdef ClockEvent ev = ClockEvent(
            self, interval, callback, timeout, 0,
            clock_ended_callback=clock_ended_callback, release_ref=release_ref)
        if release_ref:
            ev.release()
        return ev

    cpdef ClockEvent create_trigger(
            self, callback, timeout=0, interval=False, release_ref=True):
        '''Create a Trigger event. It is thread safe but not ``__del__`` or
        ``__dealloc__`` safe (see :meth:`schedule_del_safe`).
        Check module documentation for more information.

        To cancel the event before it is executed, call :meth:`ClockEvent.cancel`
        on the returned event.
        To schedule it again, simply call the event (``event()``) and it'll be safely
        rescheduled if it isn't already scheduled.

        :Parameters:

            `callback`: callable
                The callback to execute from kivy. It takes a single parameter - the
                current elapsed kivy time.
            `timeout`: float
                How long to wait before calling the callback.
            `interval`: bool
                Whether the callback should be called once (False) or repeatedly
                with a period of ``timeout`` (True) like :meth:`schedule_interval`.
            `release_ref`: bool
                If True, the default, then if ``callback``
                is a class method and the object has no references to it, then
                the object may be garbage collected and the callbacks won't be called.
                If False, the clock keeps a reference to the object preventing it
                from being garbage collected - so it will be called.

        :returns:

            A :class:`ClockEvent` instance. To schedule the callback of this
            instance, you can call it.

        .. versionadded:: 1.0.5

        .. versionchanged:: 1.10.0

            ``interval`` has been added.

        .. versionchanged:: 2.0.0

            ``release_ref`` has been added.
        '''
        cdef ClockEvent ev = ClockEvent(
            self, interval, callback, timeout, 0, release_ref=release_ref)
        if release_ref:
            ev.release()
        return ev

    cpdef schedule_lifecycle_aware_del_safe(self, callback, clock_ended_callback):
        '''Schedule a callback that is thread safe and ``__del__`` or
        ``__dealloc__`` safe similarly to :meth:`schedule_del_safe`, but the callback
        is sensitive to the clock's state.

        If this event is triggered after the clock has stopped (:meth:`stop_clock`), then a
        :class:`ClockNotRunningError` will be raised. If the error is not raised,
        then either ``callback`` or ``clock_ended_callback`` will be
        called. ``callback`` will be called when the callback
        is normally executed. If the clock is stopped before it can be executed,
        provided the app exited normally without crashing then
        ``clock_ended_callback`` will be called instead when the clock is stopped.

        :Parameters:

            `callback`: Callable
                The callback the execute from kivy. It takes no parameters and
                cannot be canceled.
            `clock_ended_callback`: callable
                A callback that will be called if the clock is stopped
                while the callback is still scheduled to be called. The callback takes
                a single parameter - the callback. If the
                app exited normally, it is guaranteed that either ``callback``
                or ``clock_ended_callback`` would have been called.

        .. versionadded:: 2.0.0
        '''
        with self._del_safe_lock:
            if self._del_safe_done:
                raise ClockNotRunningError
            self._del_queue.append((callback, clock_ended_callback))

    cpdef schedule_del_safe(self, callback):
        '''Schedule a callback that is thread safe and ``__del__`` or
        ``__dealloc__`` safe.

        It's unsafe to call various kinds of code from ``__del__`` or
        ``__dealloc__`` methods because they can be executed at any time. Most
        Kivy's Clock methods are unsafe to call the Clock from these methods. Instead,
        use this method, which is thread safe and ``__del__`` or ``__dealloc__``
        safe, to schedule the callback in the kivy thread. It'll be executed
        in order after the normal events are processed.

        :Parameters:

            `callback`: Callable
                The callback the execute from kivy. It takes no parameters and
                cannot be canceled.

        .. versionadded:: 1.11.0
        '''
        self._del_queue.append((callback, None))

    cpdef ClockEvent schedule_once(self, callback, timeout=0):
        '''Schedule an event in <timeout> seconds. If <timeout> is unspecified
        or 0, the callback will be called after the next frame is rendered.
        See :meth:`create_trigger` for advanced scheduling and more details.

        To cancel the event before it is executed, call :meth:`ClockEvent.cancel`
        on the returned event.
        If the callback is a class method, a weakref to the object is created and it
        may be garbage collected if there's no other reference to the object.

        :returns:

            A :class:`ClockEvent` instance. As opposed to
            :meth:`create_trigger` which only creates the trigger event, this
            method also schedules it.

        .. versionchanged:: 1.0.5
            If the timeout is -1, the callback will be called before the next
            frame (at :meth:`tick_draw`).
        '''
        cdef ClockEvent event
        if not callable(callback):
            raise ValueError('callback must be a callable, got %s' % callback)

        event = ClockEvent(
            self, False, callback, timeout, self._last_tick, None, True)
        return event

    cpdef ClockEvent schedule_interval(self, callback, timeout):
        '''Schedule an event to be called every <timeout> seconds.
        See :meth:`create_trigger` for advanced scheduling and more details.

        To cancel the event before it is executed, call :meth:`ClockEvent.cancel`
        on the returned event.
        If the callback is a class method, a weakref to the object is created and it
        may be garbage collected if there's no other reference to the object.

        :returns:

            A :class:`ClockEvent` instance. As opposed to
            :meth:`create_trigger` which only creates the trigger event, this
            method also schedules it.
        '''
        cdef ClockEvent event
        if not callable(callback):
            raise ValueError('callback must be a callable, got %s' % callback)

        event = ClockEvent(
            self, True, callback, timeout, self._last_tick, None, True)
        return event

    cpdef unschedule(self, callback, all=True):
        '''Remove a previously scheduled event.

        An :class:`ClockEvent` can also be canceled directly by calling
        :meth:`ClockEvent.cancel`.

        :parameters:

            `callback`: :class:`ClockEvent` or a callable.
                If it's a :class:`ClockEvent` instance, then the callback
                associated with this event will be canceled if it is
                scheduled.

                If it's a callable, then the callable will be unscheduled if it
                was scheduled.

                .. warning::

                    Passing the callback function rather than the returned
                    :class:`ClockEvent` will result in a significantly slower
                    unscheduling.
            `all`: bool
                If True and if `callback` is a callable, all instances of this
                callable will be unscheduled (i.e. if this callable was
                scheduled multiple times). Defaults to `True`.

        .. versionchanged:: 1.9.0
            The all parameter was added. Before, it behaved as if `all` was
            `True`.
        '''
        cdef ClockEvent ev, cancel
        cdef list events
        if isinstance(callback, ClockEvent):
            ev = callback
            ev.cancel()
        else:
            if all:
                events = []
                self._lock_acquire()

                ev = self._root_event
                while ev is not None:
                    if ev.get_callback() == callback:
                        events.append(ev)
                    ev = ev.next

                self._lock_release()
                for ev in events:
                    ev.cancel()
            else:
                cancel = None
                self._lock_acquire()

                ev = self._root_event
                while ev is not None:
                    if ev.get_callback() == callback:
                        cancel = ev
                        break
                    ev = ev.next
                self._lock_release()
                if cancel is not None:
                    cancel.cancel()

    cpdef _release_references(self):
        # call that function to release all the direct reference to any
        # callback and replace it with a weakref
        cdef list events = []
        cdef ClockEvent ev
        self._lock_acquire()

        ev = self._root_event
        while ev is not None:
            if ev.callback is not None and ev.release_ref:
                events.append(ev)
            ev = ev.next

        self._lock_release()
        for ev in events:
            ev.release()

    cpdef _process_del_safe_events(self):
        callbacks = self._del_queue[:]
        del self._del_queue[:len(callbacks)]
        for callback, _ in callbacks:
            try:
                callback()
            except BaseException as e:
                self.handle_exception(e)

    cpdef _process_events(self):
        cdef ClockEvent event
        cdef int done = False

        self._lock_acquire()
        if self._root_event is None:
            self._lock_release()
            self._process_del_safe_events()
            return

        self._cap_event = self._last_event
        event = self._root_event
        while not done and event is not None:
            self._next_event = event.next
            done = self._cap_event is event or self._cap_event is None
            '''Usage of _cap_event: We have to worry about this case:

            If in this iteration the cap event is canceled then at end of this
            iteration _cap_event will have shifted to current event (or to the
            event before that if current_event is done) which will not be checked
            again for being the cap event and done will never be True
            since we are passed the current event. So, when canceling,
            if _next_event is the canceled event and it's also the _cap_event
            _cap_event is set to None.
            '''

            self._lock_release()

            try:
                event.tick(self._last_tick)
            except BaseException as e:
                # cancel the event either way
                event.cancel()
                self.handle_exception(e)
            self._lock_acquire()
            event = self._next_event

        self._next_event = self._cap_event = None
        self._lock_release()

        self._process_del_safe_events()

    cpdef _process_events_before_frame(self):
        cdef ClockEvent event
        cdef int count = self.max_iteration
        cdef int found = True
        cdef int done = False

        while found:
            count -= 1
            if count == -1:
                remaining_events = '\n'.join(
                    map(str, self.get_before_frame_events()))
                Logger.critical(
                    f'Clock: Warning, too much iteration done before'
                    f' the next frame. Check your code, or increase'
                    f' the Clock.max_iteration attribute. Remaining events:\n'
                    f'{remaining_events}')
                break

            # search event that have timeout = -1
            found = False
            done = False

            self._lock_acquire()
            if self._root_event is None:
                self._lock_release()
                return

            # see _process_events for the logic
            self._cap_event = self._last_event
            event = self._root_event
            while not done and event is not None:
                if event.timeout != -1:
                    done = self._cap_event is event or self._cap_event is None
                    event = event.next
                    continue

                self._next_event = event.next
                done = self._cap_event is event or self._cap_event is None
                self._lock_release()
                found = True

                try:
                    event.tick(self._last_tick)
                except BaseException as e:
                    # cancel the event either way
                    event.cancel()
                    self.handle_exception(e)
                self._lock_acquire()
                event = self._next_event
            self._next_event = self._cap_event = None
            self._lock_release()

    cpdef get_min_timeout(self):
        '''Returns the remaining time since the start of the current frame
        for the event with the smallest timeout.
        '''
        cdef ClockEvent ev
        cdef double val = DBL_MAX
        self._lock_acquire()
        ev = self._root_event
        while ev is not None:
            if ev.timeout <= 0:
                val = 0
                break
            val = min(val, ev.timeout + ev._last_dt)
            ev = ev.next
        self._lock_release()

        return val

    cpdef get_events(self):
        '''Returns the list of :class:`ClockEvent` instances currently scheduled.
        '''
        cdef list events = []
        cdef ClockEvent ev

        self._lock_acquire()
        ev = self._root_event
        while ev is not None:
            events.append(ev)
            ev = ev.next
        self._lock_release()
        return events

    cpdef get_before_frame_events(self):
        '''Returns the list of :class:`ClockEvent` instances that are scheduled
        to be called before the next frame (``-1`` timeout).

        .. versionadded:: 2.1.0
        '''
        cdef list events = []
        cdef ClockEvent ev

        self._lock_acquire()
        ev = self._root_event
        while ev is not None:
            if ev.timeout == -1:
                events.append(ev)
            ev = ev.next
        self._lock_release()
        return events

    def handle_exception(self, e):
        """Provides an opportunity to handle an event's exception.

        If desired, the exception is handled, otherwise it should be raised
        again. By default it is raised again.

        :param e: The exception to be handled.

        .. versionadded:: 2.0.0
        """
        raise

    cpdef _process_clock_ended_del_safe_events(self):
        with self._del_safe_lock:
            self._del_safe_done = True
            callbacks = self._del_queue[:]

        del self._del_queue[:len(callbacks)]
        for callback, clock_ended_callback in callbacks:
            try:
                if clock_ended_callback is not None:
                    clock_ended_callback(callback)
            except BaseException as e:
                self.handle_exception(e)

    cpdef _process_clock_ended_callbacks(self):
        # at this point, if an event is scheduled, it should raise an exception
        cdef ClockEvent event
        cdef int done = False

        self._lock_acquire()
        if self._root_event is None:
            self._lock_release()
            self._process_clock_ended_del_safe_events()
            return

        # see _process_events for the logic
        self._cap_event = self._last_event
        event = self._root_event
        while not done and event is not None:
            self._next_event = event.next
            done = self._cap_event is event or self._cap_event is None
            self._lock_release()

            callback = event.get_clock_ended_callback()
            if callback is not None:
                try:
                    callback(event)
                except BaseException as e:
                    self.handle_exception(e)
            self._lock_acquire()
            event = self._next_event

        self._next_event = self._cap_event = None
        self._lock_release()

        self._process_clock_ended_del_safe_events()


cdef class CyClockBaseFree(CyClockBase):
    '''A clock class that supports scheduling free events in addition to normal
    events.

    Each of the :meth:`~CyClockBase.create_trigger`,
    :meth:`~CyClockBase.schedule_once`, and :meth:`~CyClockBase.schedule_interval`
    methods, which create a normal event, have a corresponding method
    for creating a free event.
    '''

    cpdef FreeClockEvent create_lifecycle_aware_trigger(
            self, callback, clock_ended_callback, timeout=0, interval=False,
            release_ref=True):
        cdef FreeClockEvent event
        if not callable(callback):
            raise ValueError('callback must be a callable, got %s' % callback)

        event = FreeClockEvent(
            False, self, interval, callback, timeout, 0,
            clock_ended_callback=clock_ended_callback, release_ref=release_ref)
        if release_ref:
            event.release()
        return event

    cpdef FreeClockEvent create_trigger(
            self, callback, timeout=0, interval=False, release_ref=True):
        cdef FreeClockEvent event
        if not callable(callback):
            raise ValueError('callback must be a callable, got %s' % callback)

        event = FreeClockEvent(
            False, self, interval, callback, timeout, 0, release_ref=release_ref)
        if release_ref:
            event.release()
        return event

    cpdef FreeClockEvent schedule_once(self, callback, timeout=0):
        cdef FreeClockEvent event
        if not callable(callback):
            raise ValueError('callback must be a callable, got %s' % callback)

        event = FreeClockEvent(
            False, self, False, callback, timeout, self._last_tick, None, True)
        return event

    cpdef FreeClockEvent schedule_interval(self, callback, timeout):
        cdef FreeClockEvent event
        if not callable(callback):
            raise ValueError('callback must be a callable, got %s' % callback)

        event = FreeClockEvent(
            False, self, True, callback, timeout, self._last_tick, None, True)
        return event

    cpdef FreeClockEvent create_lifecycle_aware_trigger_free(
            self, callback, clock_ended_callback, timeout=0, interval=False,
            release_ref=True):
        '''Similar to :meth:`~CyClockBase.create_lifecycle_aware_trigger`, but instead creates
        a free event.
        '''
        cdef FreeClockEvent event
        if not callable(callback):
            raise ValueError('callback must be a callable, got %s' % callback)

        event = FreeClockEvent(
            True, self, interval, callback, timeout, 0,
            clock_ended_callback=clock_ended_callback, release_ref=release_ref)
        if release_ref:
            event.release()
        return event

    cpdef FreeClockEvent create_trigger_free(
            self, callback, timeout=0, interval=False, release_ref=True):
        '''Similar to :meth:`~CyClockBase.create_trigger`, but instead creates
        a free event.
        '''
        cdef FreeClockEvent event
        if not callable(callback):
            raise ValueError('callback must be a callable, got %s' % callback)

        event = FreeClockEvent(
            True, self, interval, callback, timeout, 0, release_ref=release_ref)
        if release_ref:
            event.release()
        return event

    cpdef FreeClockEvent schedule_once_free(self, callback, timeout=0):
        '''Similar to :meth:`~CyClockBase.schedule_once`, but instead creates
        a free event.
        '''
        cdef FreeClockEvent event
        if not callable(callback):
            raise ValueError('callback must be a callable, got %s' % callback)

        event = FreeClockEvent(
            True, self, False, callback, timeout, self._last_tick, None, True)
        return event

    cpdef FreeClockEvent schedule_interval_free(self, callback, timeout):
        '''Similar to :meth:`~CyClockBase.schedule_interval`, but instead creates
        a free event.
        '''
        cdef FreeClockEvent event
        if not callable(callback):
            raise ValueError('callback must be a callable, got %s' % callback)

        event = FreeClockEvent(
            True, self, True, callback, timeout, self._last_tick, None, True)
        return event

    cpdef _process_free_events(self, double last_tick):
        cdef FreeClockEvent event
        cdef int done = False

        self._lock_acquire()
        if self._root_event is None:
            self._lock_release()
            return

        self._cap_event = self._last_event
        event = self._root_event
        while not done and event is not None:
            if not event.free:
                done = self._cap_event is event or self._cap_event is None
                event = event.next
                continue

            self._next_event = event.next
            done = self._cap_event is event or self._cap_event is None

            self._lock_release()

            try:
                event.tick(last_tick)
            except BaseException as e:
                # cancel the event either way
                event.cancel()
                self.handle_exception(e)
            self._lock_acquire()
            event = self._next_event

        self._next_event = self._cap_event = None
        self._lock_release()

    cpdef get_min_free_timeout(self):
        '''Returns the remaining time since the start of the current frame
        for the *free* event with the smallest timeout.
        '''
        cdef FreeClockEvent ev
        cdef double val = DBL_MAX

        self._lock_acquire()
        ev = self._root_event
        while ev is not None:
            if ev.free:
                if ev.timeout <= 0:
                    val = 0
                    break
                val = min(val, ev.timeout + ev._last_dt)
            ev = ev.next
        self._lock_release()

        return val



#cython: c_string_type=unicode, c_string_encoding=utf8
'''
Shader
======

The :class:`Shader` class handles the compilation of the vertex and fragment
shader as well as the creation of the program in OpenGL.

.. todo::

    Include more complete documentation about the shader.

Header inclusion
----------------

.. versionadded:: 1.0.7

When you are creating a Shader, Kivy will always include default parameters. If
you don't want to rewrite this each time you want to customize / write a new
shader, you can add the "$HEADER$" token and it will be replaced by the
corresponding shader header.

Here is the header for the fragment Shader:

.. include:: ../../kivy/data/glsl/header.fs
    :literal:

And the header for vertex Shader:

.. include:: ../../kivy/data/glsl/header.vs
    :literal:


Single file glsl shader programs
--------------------------------

.. versionadded:: 1.6.0

To simplify shader management, the vertex and fragment shaders can be loaded
automatically from a single glsl source file (plain text). The file should
contain sections identified by a line starting with '---vertex' and
'---fragment' respectively (case insensitive), e.g. ::

    // anything before a meaningful section such as this comment are ignored

    ---VERTEX SHADER--- // vertex shader starts here
    void main(){
        ...
    }

    ---FRAGMENT SHADER--- // fragment shader starts here
    void main(){
        ...
    }

The source property of the Shader should be set to the filename of a glsl
shader file (of the above format), e.g. `phong.glsl`
'''

__all__ = ('Shader', )

include "../include/config.pxi"
include "common.pxi"
include "gl_debug_logger.pxi"

from os.path import join

from kivy.graphics.cgl cimport *

from kivy.graphics.vertex cimport vertex_attr_t
from kivy.graphics.transformation cimport Matrix
from kivy.graphics.context cimport get_context
from kivy.logger import Logger
from kivy.cache import Cache
from kivy import kivy_shader_dir

cdef str header_vs = ''
cdef str header_fs = ''
cdef str default_vs = ''
cdef str default_fs = ''
with open(join(kivy_shader_dir, 'header.vs')) as fin:
    header_vs = fin.read()
with open(join(kivy_shader_dir, 'header.fs')) as fin:
    header_fs = fin.read()
with open(join(kivy_shader_dir, 'default.vs')) as fin:
    default_vs = fin.read()
with open(join(kivy_shader_dir, 'default.fs')) as fin:
    default_fs = fin.read()


cdef class ShaderSource:

    def __cinit__(self, shadertype):
        self.shader = -1
        self.shadertype = shadertype

    cdef set_source(self, char *source):
        cdef GLint success = 0
        cdef GLuint error, shader
        cdef str ctype, cacheid

        # XXX to ensure that shader is ok, read error state right now.
        cgl.glGetError()

        # create and compile
        shader = cgl.glCreateShader(self.shadertype)
        cgl.glShaderSource(shader, 1, <const_char_ptr*> &source, NULL)
        cgl.glCompileShader(shader)

        # show any messages
        ctype = 'vertex' if self.shadertype == GL_VERTEX_SHADER else 'fragment'

        # ensure compilation is ok
        cgl.glGetShaderiv(shader, GL_COMPILE_STATUS, &success)

        if success == GL_FALSE:
            error = cgl.glGetError()
            Logger.error('Shader: <%s> failed to compile (gl:%d)' % (
                ctype, error))
            self.process_message('%s shader' % ctype, self.get_shader_log(shader))
            cgl.glDeleteShader(shader)
            return

        Logger.debug('Shader: %s compiled successfully' % ctype.capitalize())
        self.shader = shader

    def __dealloc__(self):
        if self.shader != -1:
            get_context().dealloc_shader_source(self.shader)

    cdef int is_compiled(self):
        if self.shader != -1:
            return 1
        return 0

    cdef void process_message(self, str ctype, message):
        message = message.strip()
        if message:
            Logger.info('Shader: %s: <%s>' % (ctype, message))

    cdef get_shader_log(self, int shader):
        '''Return the shader log.
        '''
        cdef char *msg
        cdef bytes py_msg
        cdef int info_length
        cgl.glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &info_length)
        if info_length <= 0:
            return ""
        msg = <char *>malloc(info_length * sizeof(char))
        if msg == NULL:
            return ""
        msg[0] = b"\0"
        cgl.glGetShaderInfoLog(shader, info_length, NULL, msg)
        py_msg = msg
        free(msg)
        return py_msg


cdef class Shader:
    '''Create a vertex or fragment shader.

    :Parameters:
        `vs`: string, defaults to None
            Source code for vertex shader
        `fs`: string, defaults to None
            Source code for fragment shader
    '''
    def __cinit__(self):
        self._success = 0
        self.program = 0
        self.vertex_shader = None
        self.fragment_shader = None
        self.uniform_locations = dict()
        self.uniform_values = dict()

    def __init__(self, str vs=None, str fs=None, str source=None):
        self.program = cgl.glCreateProgram()
        if source:
            self.source = source
        else:
            self._source = None
            self.fs = fs
            self.vs = vs

    def __dealloc__(self):
        get_context().dealloc_shader(self)

    cdef void reload(self):
        # Note that we don't free previous created shaders. The current reload
        # is called only when the gl context is reset. If we do it, we might
        # free newly created shaders (id collision)
        cgl.glUseProgram(0)

        # avoid shaders to be collected
        if self.vertex_shader:
            self.vertex_shader.shader = -1
            self.vertex_shader = None
        if self.fragment_shader:
            self.fragment_shader.shader = -1
            self.fragment_shader = None

        #self.uniform_values = dict()
        self.uniform_locations = dict()
        self._success = 0
        self._current_vertex_format = None
        self.program = cgl.glCreateProgram()
        self.fs = self.fs
        self.vs = self.vs

    cdef void use(self):
        '''Use the shader.
        '''
        cgl.glUseProgram(self.program)
        log_gl_error('Shader.use-glUseProgram')
        for k, v in self.uniform_values.iteritems():
            self.upload_uniform(k, v)
        if cgl_get_initialized_backend_name() == 'glew':
            # XXX Very very weird bug. On virtualbox / win7 / glew, if we don't call
            # glFlush or glFinish or glGetIntegerv(GL_CURRENT_PROGRAM, ...), it seem
            # that the pipeline is broken, and we have glitch issue. In order to
            # prevent that on possible other hardware, i've (mathieu) preferred to
            # include a glFlush here. However, it could be nice to know exactly what
            # is going on. Even the glGetIntegerv() is not working here. Broken
            # driver on virtualbox / win7 ????
            # FIXME maybe include that instruction for glew usage only.
            cgl.glFlush()

    cdef void stop(self):
        '''Stop using the shader.
        '''
        cgl.glUseProgram(0)
        log_gl_error('Shader.stop-glUseProgram')

    cdef int set_uniform(self, str name, value) except -1:
        if name in self.uniform_values and self.uniform_values[name] == value:
            return 0
        cdef GLint data
        cgl.glGetIntegerv(GL_CURRENT_PROGRAM, &data)
        log_gl_error('Shader.set_uniform-glGetIntegerv')
        if data != self.program:
            cgl.glUseProgram(self.program)
            log_gl_error('Shader.set_uniform-glUseProgram')
        self.uniform_values[name] = value
        self.upload_uniform(name, value)
        return 0

    cdef int upload_uniform(self, str name, value) except -1:
        '''Pass a uniform variable to the shader.
        '''
        cdef long vec_size, index, x, y
        cdef int list_size
        cdef int loc, i1, i2, i3, i4
        cdef float f1, f2, f3, f4
        cdef tuple tuple_value
        cdef list list_value
        cdef GLfloat *float_list
        cdef GLint *int_list
        val_type = type(value)
        loc = self.uniform_locations.get(name, -1)
        if loc == -1:
            loc = self.get_uniform_loc(name)

        #Logger.debug('Shader: uploading uniform %s (loc=%d, value=%r)' % (name, loc, value))
        if loc == -1:
            #Logger.debug('Shader: -> ignored')
            return 0
        #Logger.debug('Shader: -> (gl:%d) %s' % (glGetError(), str(value)))

        if val_type is Matrix:
            self.upload_uniform_matrix(loc, value)
            log_gl_error('Shader.upload_uniform-glUniformMatrix4fv'
                ' {name}'.format(name=name))
        elif val_type is int:
            cgl.glUniform1i(loc, value)
            log_gl_error('Shader.upload_uniform-glUniform1i'
                ' {name}'.format(name=name))
        elif val_type is float:
            cgl.glUniform1f(loc, value)
            log_gl_error('Shader.upload_uniform-glUniform1f'
                ' {name}'.format(name=name))
        elif val_type is list:
            list_value = value
            val_type = type(list_value[0])
            vec_size = <long>len(list_value)
            if val_type is float:
                if vec_size == 2:
                    f1, f2 = list_value
                    cgl.glUniform2f(loc, f1, f2)
                    log_gl_error('Shader.upload_uniform-glUniform2f'
                        ' {name}'.format(name=name))
                elif vec_size == 3:
                    f1, f2, f3 = list_value
                    cgl.glUniform3f(loc, f1, f2, f3)
                    log_gl_error('Shader.upload_uniform-glUniform3f'
                        ' {name}'.format(name=name))
                elif vec_size == 4:
                    f1, f2, f3, f4 = list_value
                    cgl.glUniform4f(loc, f1, f2, f3, f4)
                    log_gl_error('Shader.upload_uniform-glUniform4f'
                        ' {name}'.format(name=name))
                else:
                    float_list = <GLfloat *>malloc(vec_size * sizeof(GLfloat))
                    if float_list is NULL:
                        raise MemoryError()
                    for index in xrange(vec_size):
                        float_list[index] = <GLfloat>list_value[index]
                    cgl.glUniform1fv(loc, <GLint>vec_size, float_list)
                    log_gl_error('Shader.upload_uniform-glUniform1fv'
                        ' {name}'.format(name=name))
                    free(float_list)
            elif val_type is int:
                if vec_size == 2:
                    i1, i2 = list_value
                    cgl.glUniform2i(loc, i1, i2)
                    log_gl_error('Shader.upload_uniform-glUniform2i'
                        ' {name}'.format(name=name))
                elif vec_size == 3:
                    i1, i2, i3 = list_value
                    cgl.glUniform3i(loc, i1, i2, i3)
                    log_gl_error('Shader.upload_uniform-glUniform3i'
                        ' {name}'.format(name=name))
                elif vec_size == 4:
                    i1, i2, i3, i4 = list_value
                    cgl.glUniform4i(loc, i1, i2, i3, i4)
                    log_gl_error('Shader.upload_uniform-glUniform4i'
                        ' {name}'.format(name=name))
                else:
                    int_list = <int *>malloc(vec_size * sizeof(GLint))
                    if int_list is NULL:
                        raise MemoryError()
                    for index in xrange(vec_size):
                        int_list[index] = <GLint>list_value[index]
                    cgl.glUniform1iv(loc, <GLint>vec_size, int_list)
                    log_gl_error('Shader.upload_uniform-glUniform1iv'
                        ' {name}'.format(name=name))
                    free(int_list)
            elif val_type is list:
                list_size = <int>len(value)
                vec_size = <long>len(value[0])
                val_type = type(value[0][0])
                if val_type is float:
                    float_list = <GLfloat *>malloc(
                            list_size * vec_size * sizeof(GLfloat))
                    if float_list is NULL:
                        raise MemoryError()
                    for x in xrange(list_size):
                        for y in xrange(vec_size):
                            float_list[vec_size * x + y] = <GLfloat>value[x][y]
                    if vec_size == 2:
                        cgl.glUniform2fv(loc, list_size, float_list)
                        log_gl_error('Shader.upload_uniform-glUniform2fv'
                            ' {name}'.format(name=name))
                    elif vec_size == 3:
                        cgl.glUniform3fv(loc, list_size, float_list)
                        log_gl_error('Shader.upload_uniform-glUniform3fv'
                            ' {name}'.format(name=name))
                    elif vec_size == 4:
                        cgl.glUniform4fv(loc, list_size, float_list)
                        log_gl_error('Shader.upload_uniform-glUniform4fv'
                            ' {name}'.format(name=name))
                    else:
                        Logger.debug(
                            'Shader: unsupported {}x{} float array'.format(
                            list_size, vec_size))
                    free(float_list)
                elif val_type is int:
                    int_list = <GLint *>malloc(
                            list_size * vec_size * sizeof(GLint))
                    if int_list is NULL:
                        raise MemoryError()
                    for x in xrange(list_size):
                        for y in xrange(vec_size):
                            int_list[vec_size * x + y] = <GLint>value[x][y]
                    if vec_size == 2:
                        cgl.glUniform2iv(loc, list_size, int_list)
                        log_gl_error('Shader.upload_uniform-glUniform2iv'
                            ' {name}'.format(name=name))
                    elif vec_size == 3:
                        cgl.glUniform3iv(loc, list_size, int_list)
                        log_gl_error('Shader.upload_uniform-glUniform3iv'
                            ' {name}'.format(name=name))
                    elif vec_size == 4:
                        cgl.glUniform4iv(loc, list_size, int_list)
                        log_gl_error('Shader.upload_uniform-glUniform4iv'
                            ' {name}'.format(name=name))
                    else:
                        Logger.debug(
                            'Shader: unsupported {}x{} int array'.format(
                            list_size, vec_size))
                    free(int_list)
        elif val_type is tuple:
            tuple_value = value
            val_type = type(tuple_value[0])
            vec_size = <long>len(tuple_value)
            if val_type is float:
                if vec_size == 2:
                    f1, f2 = tuple_value
                    cgl.glUniform2f(loc, f1, f2)
                    log_gl_error('Shader.upload_uniform-glUniform2f'
                        ' {name}'.format(name=name))
                elif vec_size == 3:
                    f1, f2, f3 = tuple_value
                    cgl.glUniform3f(loc, f1, f2, f3)
                    log_gl_error('Shader.upload_uniform-glUniform3f'
                        ' {name}'.format(name=name))
                elif vec_size == 4:
                    f1, f2, f3, f4 = tuple_value
                    cgl.glUniform4f(loc, f1, f2, f3, f4)
                    log_gl_error('Shader.upload_uniform-glUniform4f'
                        ' {name}'.format(name=name))
            elif val_type is int:
                if vec_size == 2:
                    i1, i2 = tuple_value
                    cgl.glUniform2i(loc, i1, i2)
                    log_gl_error('Shader.upload_uniform-glUniform2i'
                        ' {name}'.format(name=name))
                elif vec_size == 3:
                    i1, i2, i3 = tuple_value
                    cgl.glUniform3i(loc, i1, i2, i3)
                    log_gl_error('Shader.upload_uniform-glUniform3i'
                        ' {name}'.format(name=name))
                elif vec_size == 4:
                    i1, i2, i3, i4 = tuple_value
                    cgl.glUniform4i(loc, i1, i2, i3, i4)
                    log_gl_error('Shader.upload_uniform-glUniform4i'
                        ' {name}'.format(name=name))
            elif val_type is list:
                list_size = <int>len(value)
                vec_size = <long>len(value[0])
                val_type = type(value[0][0])
                if val_type is float:
                    float_list = <GLfloat *>malloc(
                            list_size * vec_size * sizeof(GLfloat))
                    if float_list is NULL:
                        raise MemoryError()
                    for x in xrange(list_size):
                        for y in xrange(vec_size):
                            float_list[vec_size * x + y] = <GLfloat>value[x][y]
                    if vec_size == 2:
                        cgl.glUniform2fv(loc, list_size, float_list)
                        log_gl_error('Shader.upload_uniform-glUniform2fv'
                            ' {name}'.format(name=name))
                    elif vec_size == 3:
                        cgl.glUniform3fv(loc, list_size, float_list)
                        log_gl_error('Shader.upload_uniform-glUniform3fv'
                            ' {name}'.format(name=name))
                    elif vec_size == 4:
                        cgl.glUniform4fv(loc, list_size, float_list)
                        log_gl_error('Shader.upload_uniform-glUniform4fv'
                            ' {name}'.format(name=name))
                    else:
                        Logger.debug(
                            'Shader: unsupported {}x{} float array'.format(
                            list_size, vec_size))
                    free(float_list)
                elif val_type is int:
                    int_list = <GLint *>malloc(
                            list_size * vec_size * sizeof(GLint))
                    if int_list is NULL:
                        raise MemoryError()
                    for x in xrange(list_size):
                        for y in xrange(vec_size):
                            int_list[vec_size * x + y] = <GLint>value[x][y]
                    if vec_size == 2:
                        cgl.glUniform2iv(loc, list_size, int_list)
                        log_gl_error('Shader.upload_uniform-glUniform2iv'
                            ' {name}'.format(name=name))
                    elif vec_size == 3:
                        cgl.glUniform3iv(loc, list_size, int_list)
                        log_gl_error('Shader.upload_uniform-glUniform3iv'
                            ' {name}'.format(name=name))
                    elif vec_size == 4:
                        cgl.glUniform4iv(loc, list_size, int_list)
                        log_gl_error('Shader.upload_uniform-glUniform4iv'
                            ' {name}'.format(name=name))
                    else:
                        Logger.debug(
                            'Shader: unsupported {}x{} int array'.format(
                            list_size, vec_size))
                    free(int_list)
        else:
            raise Exception('for <%s>, type not handled <%s>' % (name, val_type))
        return 0

    cdef void upload_uniform_matrix(self, int loc, Matrix value):
        cdef GLfloat mat[16]
        for x in xrange(16):
            mat[x] = <GLfloat>value.mat[x]
        cgl.glUniformMatrix4fv(loc, 1, False, mat)

    cdef int get_uniform_loc(self, str name) except *:
        cdef bytes c_name = name.encode('utf-8')
        cdef int loc = cgl.glGetUniformLocation(self.program, c_name)
        log_gl_error(
            'Shader.get_uniform_loc-glGetUniformLocation ({name})'.format(
            name=name))
        self.uniform_locations[name] = loc
        return loc

    cdef void bind_vertex_format(self, VertexFormat vertex_format):
        cdef int i
        cdef vertex_attr_t *attr
        cdef bytes name

        # if the current vertex format used in the shader is the current one, do
        # nothing.
        # the same vertex format might be used by others shaders, so the
        # attr.index would not be accurate. we need to update it as well.
        if vertex_format and self._current_vertex_format is vertex_format and \
                vertex_format.last_shader is self:
            return

        # unbind the previous vertex format
        if self._current_vertex_format:
            for i in range(self._current_vertex_format.vattr_count):
                attr = &self._current_vertex_format.vattr[i]
                if attr.per_vertex == 0:
                    continue
                if attr.index != <unsigned int>-1:
                  cgl.glDisableVertexAttribArray(attr.index)
                log_gl_error(
                    'Shader.bind_vertex_format-glDisableVertexAttribArray')

        # bind the new vertex format
        if vertex_format:
            vertex_format.last_shader = self
            for i in range(vertex_format.vattr_count):
                attr = &vertex_format.vattr[i]
                if attr.per_vertex == 0:
                    continue
                name = <bytes>attr.name
                attr.index = cgl.glGetAttribLocation(self.program, <char *>name)
                if attr.index != <unsigned int>-1:
                    cgl.glEnableVertexAttribArray(attr.index)
                log_gl_error(
                    'Shader.bind_vertex_format-glEnableVertexAttribArray')

        # save for the next run.
        self._current_vertex_format = vertex_format

    cdef int build(self) except -1:
        self.build_vertex()
        self.build_fragment()
        return 0

    cdef int build_vertex(self, int link=1) except -1:
        if self.vertex_shader is not None:
            cgl.glDetachShader(self.program, self.vertex_shader.shader)
            log_gl_error('Shader.build_vertex-glDetachShader')
            self.vertex_shader = None
        self.vertex_shader = self.compile_shader(self.vert_src, GL_VERTEX_SHADER)
        if self.vertex_shader is not None:
            cgl.glAttachShader(self.program, self.vertex_shader.shader)
            log_gl_error('Shader.build_vertex-glAttachShader')
        if link:
            self.link_program()
        return 0

    cdef int build_fragment(self, int link=1) except -1:
        if self.fragment_shader is not None:
            cgl.glDetachShader(self.program, self.fragment_shader.shader)
            log_gl_error('Shader.build_fragment-glDetachShader')
            self.fragment_shader = None
        self.fragment_shader = self.compile_shader(self.frag_src, GL_FRAGMENT_SHADER)
        if self.fragment_shader is not None:
            cgl.glAttachShader(self.program, self.fragment_shader.shader)
            log_gl_error('Shader.build_fragment-glAttachShader')
        if link:
            self.link_program()

    cdef int link_program(self) except -1:
        if self.vertex_shader is None or self.fragment_shader is None:
            return 0

        # XXX to ensure that shader is ok, read error state right now.
        cgl.glGetError()

        cgl.glLinkProgram(self.program)
        self.uniform_locations = dict()
        error = cgl.glGetError()
        if error:
            Logger.error('Shader: GL error %d' % error)
        if not self.is_linked():
            self._success = 0
            self.process_message('program', self.get_program_log(self.program))
            raise Exception('Shader did not link, check info log.')
        self._success = 1
        return 0

    cdef int is_linked(self):
        cdef GLint result = 0
        cgl.glGetProgramiv(self.program, GL_LINK_STATUS, &result)
        if result == GL_TRUE:
            return 1
        return 0

    cdef ShaderSource compile_shader(self, str source, int shadertype):
        cdef ShaderSource shader
        cdef str ctype, cacheid
        cdef bytes b_source = source.encode('utf-8')

        ctype = 'vertex' if shadertype == GL_VERTEX_SHADER else 'fragment'

        # try to check if the shader exist in the Cache first
        cacheid = '%s|%s' % (ctype, source)
        shader = Cache.get('kv.shader', cacheid)
        if shader is not None:
            return shader

        shader = ShaderSource(shadertype)
        shader.set_source(b_source)
        if shader.is_compiled() == 0:
            self._success = 0
            return None

        Cache.append('kv.shader', cacheid, shader)
        return shader

    cdef get_program_log(self, shader):
        '''Return the program log.'''
        cdef char msg[2048]
        cdef GLsizei length
        msg[0] = b'\0'
        cgl.glGetProgramInfoLog(shader, 2048, &length, msg)
        # XXX don't use the msg[:length] as a string directly, or the unicode
        # will fail on shitty driver. Ie, some Intel drivers return a static
        # uninitialized string of length 40, with just a content of "Success.\n\0"
        # Trying to decode data after \0 will just fail. So use bytes, and
        # convert only the part before \0.
        # XXX Also, we cannot use directly msg as a python string, as some
        # others drivers doesn't include a \0 (which is great.)
        if length == 0:
            return ""
        cdef bytes ret = msg[:length]
        return ret.split(b'\0')[0].decode('utf-8')

    cdef void process_message(self, str ctype, message):
        message = message.strip()
        if message:
            Logger.info('Shader: %s: <%s>' % (ctype, message))

    #
    # Python access
    #

    @property
    def source(self):
        '''glsl  source code.

        source should be the filename of a glsl shader that contains both the
        vertex and fragment shader sourcecode, each designated by a section
        header consisting of one line starting with either "--VERTEX" or
        "--FRAGMENT" (case insensitive).

        .. versionadded:: 1.6.0
        '''
        return self._source

    @source.setter
    def source(self, object source):
        self._source = source
        if source is None:
            self.vs = None
            self.fs = None
            return
        self.vert_src = ""
        self.frag_src = ""
        glsl_source = "\n"
        Logger.info('Shader: Read <{}>'.format(self._source))
        with open(self._source) as fin:
            glsl_source += fin.read()
        sections = glsl_source.split('\n---')
        for section in sections:
            lines = section.split('\n')
            if lines[0].lower().startswith("vertex"):
                _vs = '\n'.join(lines[1:])
                self.vert_src = _vs.replace('$HEADER$', header_vs)
            if lines[0].lower().startswith("fragment"):
                _fs = '\n'.join(lines[1:])
                self.frag_src = _fs.replace('$HEADER$', header_fs)
        self.build_vertex(0)
        self.build_fragment(0)
        self.link_program()

    @property
    def vs(self):
        '''Vertex shader source code.

        If you set a new vertex shader code source, it will be automatically
        compiled and will replace the current vertex shader.
        '''
        return self.vert_src

    @vs.setter
    def vs(self, object source):
        if source is None:
            source = default_vs
        source = source.replace('$HEADER$', header_vs)
        self.vert_src = source
        self.build_vertex()

    @property
    def fs(self):
        '''Fragment shader source code.

        If you set a new fragment shader code source, it will be automatically
        compiled and will replace the current fragment shader.
        '''
        return self.frag_src

    @fs.setter
    def fs(self, object source):
        if source is None:
            source = default_fs
        source = source.replace('$HEADER$', header_fs)
        self.frag_src = source
        self.build_fragment()

    @property
    def success(self):
        '''Indicate whether the shader loaded successfully and is ready for
        usage or not.
        '''
        return self._success



"""
Weak Proxy
==========

In order to allow garbage collection, the weak proxy provides
`weak references <https://en.wikipedia.org/wiki/Weak_reference>`_ to objects.
It effectively enhances the
`weakref.proxy <https://docs.python.org/2/library/weakref.html#weakref.proxy>`_
by adding comparison support.
"""

include "include/config.pxi"
import weakref
import operator


cdef class WeakProxy(object):
    '''Replacement for weakref.proxy to support comparisons
    '''
    cdef public object __ref

    def __init__(self, obj, destructor=None):
        self.__ref = weakref.ref(obj, destructor)

    def __ref__(self):
        r = self.__ref()
        if r is None:
            raise ReferenceError('weakly-referenced object no longer exists')
        return r

    def __getattr__(self, name):
        return getattr(self.__ref__(), name)

    def __setattr__(self, name, value):
        setattr(self.__ref__(), name, value)

    def __delattr__(self, name):
        delattr(self.__ref__(), name)

    @property
    def __class__(self):
        return self.__ref__().__class__

    def __dir__(self):
        r = self.__ref()
        if not r:
            return []
        return dir(r)

    def __reversed__(self):
        return reversed(self.__ref__())

    def __round__(self):
        return round(self.__ref__())

    def __hash__(self):
        return hash(self.__ref__())

    def __richcmp__(self, other, op):
        r = self.__ref__()
        if op == 0:
            return r < other
        elif op == 1:
            return r <= other
        elif op == 2:
            return r == other
        elif op == 3:
            return r != other
        elif op == 4:
            return r > other
        elif op == 5:
            return r >= other

    def __nonzero__(self):
        return bool(self.__ref__())

    if not PY3:
        def __bool__(self):
            return bool(self.__ref__())

    def __add__(self, other):
        return self.__ref__() + other

    def __sub__(self, other):
        return self.__ref__() - other

    def __mul__(self, other):
        return self.__ref__() * other

    def __div__(self, other):
        return operator.div(self.__ref__(), other)

    def __truediv__(self, other):
        return operator.truediv(self.__ref__(), other)

    def __floordiv__(self, other):
        return self.__ref__() // other

    def __mod__(self, other):
        return self.__ref__() % other

    def __divmod__(self, other):
        return divmod(self.__ref__(), other)

    def __pow__(self, other, modulo):
        return pow(self.__ref__(), other, modulo)

    def __lshift__(self, other):
        return self.__ref__() << other

    def __rshift__(self, other):
        return self.__ref__() >> other

    def __and__(self, other):
        return self.__ref__() & other

    def __xor__(self, other):
        return self.__ref__() ^ other

    def __or__(self, other):
        return self.__ref__() | other

    def __radd__(self, other):
        return other + self.__ref__()

    def __rsub__(self, other):
        return other - self.__ref__()

    def __rmul__(self, other):
        return other * self.__ref__()

    def __rdiv__(self, other):
        return operator.div(other, self.__ref__())

    def __rtruediv__(self, other):
        return operator.truediv(other, self.__ref__())

    def __rfloordiv__(self, other):
        return other // self.__ref__()

    def __rmod__(self, other):
        return other % self.__ref__()

    def __rdivmod__(self, other):
        return divmod(other, self.__ref__())

    def __rpow__(self, other, modulo):
        return pow(other, self.__ref__(), modulo)

    def __rlshift__(self, other):
        return other << self.__ref__()

    def __rrshift__(self, other):
        return other >> self.__ref__()

    def __rand__(self, other):
        return other & self.__ref__()

    def __rxor__(self, other):
        return other ^ self.__ref__()

    def __ror__(self, other):
        return other | self.__ref__()

    def __iadd__(self, other):
        r = self.__ref__()
        r += other
        self.__ref = weakref.ref(r)
        return self

    def __isub__(self, other):
        r = self.__ref__()
        r -= other
        self.__ref = weakref.ref(r)
        return self

    def __imul__(self, other):
        r = self.__ref__()
        r *= other
        self.__ref = weakref.ref(r)
        return self

    def __idiv__(self, other):
        r = operator.idiv(self.__ref__(), other)
        self.__ref = weakref.ref(r)
        return self

    def __itruediv__(self, other):
        r = operator.itruediv(self.__ref__(), other)
        self.__ref = weakref.ref(r)
        return self

    def __ifloordiv__(self, other):
        r = self.__ref__()
        r //= other
        self.__ref = weakref.ref(r)
        return self

    def __imod__(self, other):
        r = self.__ref__()
        r %= other
        self.__ref = weakref.ref(r)
        return self

    def __ipow__(self, other):
        r = self.__ref__()
        r **= other
        self.__ref = weakref.ref(r)
        return self

    def __ilshift__(self, other):
        r = self.__ref__()
        r <<= other
        self.__ref = weakref.ref(r)
        return self

    def __irshift__(self, other):
        r = self.__ref__()
        r >>= other
        self.__ref = weakref.ref(r)
        return self

    def __iand__(self, other):
        r = self.__ref__()
        r &= other
        self.__ref = weakref.ref(r)
        return self

    def __ixor__(self, other):
        r = self.__ref__()
        r ^= other
        self.__ref = weakref.ref(r)
        return self

    def __ior__(self, other):
        r = self.__ref__()
        r |= other
        self.__ref = weakref.ref(r)
        return self

    def __neg__(self):
        return -self.__ref__()

    def __pos__(self):
        return +self.__ref__()

    def __abs__(self):
        return abs(self.__ref__())

    def __invert__(self):
        return ~self.__ref__()

    def __int__(self):
        return int(self.__ref__())

    def __long__(self):
        return long(self.__ref__())

    def __float__(self):
        return float(self.__ref__())

    def __oct__(self):
        return oct(self.__ref__())

    def __hex__(self):
        return hex(self.__ref__())

    def __index__(self):
        return operator.index(self.__ref__())

    def __len__(self):
        return len(self.__ref__())

    def __contains__(self, value):
        return value in self.__ref__()

    def __getitem__(self, key):
        return self.__ref__()[key]

    def __setitem__(self, key, value):
        self.__ref__()[key] = value

    def __delitem__(self, key):
        del self.__ref__()[key]

    def __enter__(self):
        return self.__ref__().__enter__()

    def __exit__(self, *args, **kwargs):
        return self.__ref__().__exit__(*args, **kwargs)

    def __iter__(self):
        return iter(self.__ref__())

    def __bytes__(self):
        return bytes(self.__ref())

    def __str__(self):
        return str(self.__ref())

    def __unicode__(self):
        return unicode(self.__ref())

    def __repr__(self):
        return '<WeakProxy to {!r}>'.format(self.__ref())

